#include "MMRNode.h"
#include "../../Utilities/Logger.h"
#include <sstream>
#include <QFileDialog>
#include <QDateTime>
#include <algorithm>

#include "../../Labels/StringLabelValue.h"
#include "../../Labels/ScalarLabelValue.h"

MMRNode::MMRNode(QWidget* parent) : QDialog(parent),  INode()
{
    m_specimens = NULL;
    m_additionalVariables = NULL;

    m_regressionCoefficients = new Loadings();
    m_predictedValues = new Specimens();
    m_residuals = new Specimens();
    m_RSqr = new Variables();
    m_standardisedSpecimens = new Specimens();

    setupUi(this);

    listWidget->setSelectionMode(QAbstractItemView::MultiSelection);
    listWidget_2->setSelectionMode(QAbstractItemView::MultiSelection);

    QObject::connect(radioButton, SIGNAL(clicked()), this, SLOT(leftRadioChanged()));
    QObject::connect(radioButton_2, SIGNAL(clicked()), this, SLOT(leftRadioChanged()));
    QObject::connect(radioButton_3, SIGNAL(clicked()), this, SLOT(leftRadioChanged()));
    QObject::connect(radioButton_4, SIGNAL(clicked()), this, SLOT(rightRadioChanged()));
    QObject::connect(radioButton_5, SIGNAL(clicked()), this, SLOT(rightRadioChanged()));
    QObject::connect(radioButton_6, SIGNAL(clicked()), this, SLOT(rightRadioChanged()));
    QObject::connect(listWidget, SIGNAL(itemSelectionChanged()), this, SLOT(controlsChanged()));
    QObject::connect(listWidget_2, SIGNAL(itemSelectionChanged()), this, SLOT(controlsChanged()));
    QObject::connect(lineEdit, SIGNAL(textChanged(const QString&)), this, SLOT(controlsChanged()));

    QObject::connect(comboBox, SIGNAL(activated(int)), this, SLOT(comboChanged()));
    QObject::connect(lineEdit_2, SIGNAL(editingFinished()), this, SLOT(standardise()));
    QObject::connect(pushButton, SIGNAL(clicked()), this, SLOT(browseOutputFile()));


    leftRadioChanged();
    rightRadioChanged();

    lineEdit_2->setText("0.0");
    m_updated = false;
}

MMRNode::~MMRNode()
{
}

void MMRNode::Regress()
{
    if ( radioButton_3->isChecked() &&  radioButton_6->isChecked() )
    {
        if ( listWidget->selectedItems().empty() && listWidget_2->selectedItems().empty() )
        {
            Logger::getInstance()->log("[Regression Node] No variables selected for independent or dependent variables.", Logger::WARNING);
            return;
        }
    }
    if (radioButton_3->isChecked() && listWidget->selectedItems().empty())
    {
        Logger::getInstance()->log("[Regression Node] No variables selected for dependent variables.", Logger::WARNING);
        return;
    }
    if (radioButton_6->isChecked() && listWidget_2->selectedItems().empty())
    {
        Logger::getInstance()->log("[Regression Node] No variables selected for independent variables.", Logger::WARNING);
        return;
    }

    if (m_Y.GetRows() != m_X.GetRows())
    {
        Logger::getInstance()->log("[Regression Node] Number of observations for independent and dependent data must be the same.", Logger::WARNING);
        return;
    }

    int numindependent = m_X.GetCols();
    int numdependent = m_Y.GetCols();
    int rows = m_Y.GetRows();

    m_mmr.LoadData(m_X.GetMatrix(),m_Y.GetMatrix(),numindependent,numdependent,rows);

    bool constantIsZero = false;
    m_mmr.PerformRegression(constantIsZero);

    m_mmr.GetBeta(m_regressionCoefficients);

    Matrix<double> predictions;
    Matrix<double> residuals;
    m_mmr.GetPredictions(&predictions);
    m_mmr.GetResiduals(&residuals);

    m_predictedValues->clear();
    m_residuals->clear();
    m_standardisedSpecimens->clear();

    int landmarks = m_specimens->getLandmarkCount();
    int dimensions = m_specimens->getLandmarkDimensions();
    int index = 0;
    double x0 = lineEdit_2->text().toDouble();

    if (comboBox->currentIndex() > 0)
    {
        index = comboBox->currentIndex();
    }

    for (unsigned int i=0;i<m_specimens->getSize();i++)
    {
        LandmarkSet pred(landmarks,dimensions);
        LandmarkSet res(landmarks,dimensions);
        LandmarkSet stand(landmarks,dimensions);
        LandmarkSet *lmks = m_specimens->getLandmarkSet(i);

        for (int j=0;j<landmarks;j++)
        {
            for (int k=0;k<dimensions;k++)
            {
                pred[j][k] = predictions[i][j*dimensions+k];
                res[j][k] = residuals[i][j*dimensions+k];
                double beta = (*m_regressionCoefficients)[index+1][j*dimensions+k];
                stand[j][k] = (*lmks)[j][k] - beta * (m_X[index][i] - x0);
            }
        }
        m_predictedValues->addLandmarkSet(new LandmarkSet(pred));
        m_residuals->addLandmarkSet(new LandmarkSet(res));
        m_standardisedSpecimens->addLandmarkSet(new LandmarkSet(stand));
    }

    m_predictedValues->setSpecimenLabels(m_specimens->getSpecimenLabels());
    m_predictedValues->setLandmarkLabels(m_specimens->getLandmarkLabels());
    m_residuals->setSpecimenLabels(m_specimens->getSpecimenLabels());
    m_residuals->setLandmarkLabels(m_specimens->getLandmarkLabels());
    m_standardisedSpecimens->setSpecimenLabels(m_specimens->getSpecimenLabels());
    m_standardisedSpecimens->setLandmarkLabels(m_specimens->getLandmarkLabels());

//    m_independent->clearSpecimenGroups();
//    m_dependent->clearSpecimenGroups();

    for (unsigned int i=0; i<m_specimens->getGroupNumber(); ++i)    //Add Groups
    {
        SpecimenGroup* sg = m_specimens->getGroup(i);
        SpecimenGroup* newGroup = new SpecimenGroup;
        SpecimenGroup* newGroup2 = new SpecimenGroup;
        SpecimenGroup* newGroup3 = new SpecimenGroup;
        if (sg!=0)
        {
//            m_independent->addSpecimenGroup(m_specimens->getGroupName(i), m_specimens->getGroup(i));
//            m_dependent->addSpecimenGroup(m_specimens->getGroupName(i), m_specimens->getGroup(i));
        }
        for (unsigned int j=0; j<sg->getSize(); ++j)
        {
            unsigned int specimenIndex = sg->getMember(j)->getSpecimenIndex();
            if (specimenIndex == m_specimens->getSize())
            {
                newGroup->addMean(m_specimens->getSize(),
                                  sg->getMemberID(j),
                                  sg->getMemberColor(j),
                                  sg->getMemberViewerSize(j),
                                  sg->getMemberPlotSymbol(j));
                newGroup2->addMean(m_specimens->getSize(),
                                   sg->getMemberID(j),
                                   sg->getMemberColor(j),
                                   sg->getMemberViewerSize(j),
                                   sg->getMemberPlotSymbol(j));
                newGroup3->addMean(m_specimens->getSize(),
                                   sg->getMemberID(j),
                                   sg->getMemberColor(j),
                                   sg->getMemberViewerSize(j),
                                   sg->getMemberPlotSymbol(j));
            }
            else
            {
                newGroup->addMember(  m_predictedValues->getLandmarkSet(specimenIndex),
                                      sg->getMemberID(j),
                                      sg->getMemberColor(j),
                                      sg->getMemberViewerSize(j),
                                      sg->getMemberPlotSymbol(j));
                newGroup2->addMember( m_residuals->getLandmarkSet(specimenIndex),
                                      sg->getMemberID(j),
                                      sg->getMemberColor(j),
                                      sg->getMemberViewerSize(j),
                                      sg->getMemberPlotSymbol(j));
                newGroup3->addMember( m_standardisedSpecimens->getLandmarkSet(specimenIndex),
                                      sg->getMemberID(j),
                                      sg->getMemberColor(j),
                                      sg->getMemberViewerSize(j),
                                      sg->getMemberPlotSymbol(j));
            }
        }
        newGroup->setGroupColor(sg->getGroupColor());
        newGroup->setGroupLMKSize(sg->getGroupLMKSize());
        newGroup->setGroupPlotSymbol(sg->getGroupPlotSymbol());
        newGroup2->setGroupColor(sg->getGroupColor());
        newGroup2->setGroupLMKSize(sg->getGroupLMKSize());
        newGroup2->setGroupPlotSymbol(sg->getGroupPlotSymbol());
        newGroup3->setGroupColor(sg->getGroupColor());
        newGroup3->setGroupLMKSize(sg->getGroupLMKSize());
        newGroup3->setGroupPlotSymbol(sg->getGroupPlotSymbol());

        m_predictedValues->addGroup(m_specimens->getGroupName(i), newGroup);
        m_residuals->addGroup(m_specimens->getGroupName(i), newGroup2);
        m_standardisedSpecimens->addGroup(m_specimens->getGroupName(i), newGroup3);
    }
    m_predictedValues->initialize("Regressed Specimens");
    m_residuals->initialize("Regression Residuals");
    m_standardisedSpecimens->initialize("Standardised Specimens");

    vector<double> rsqr;
    m_mmr.GetRSqr(&rsqr);
    m_RSqr->ReSize(rsqr.size(),1);
    for (unsigned int i=0;i<rsqr.size();i++)
    {
        (*m_RSqr)[i][0] = rsqr[i];
    }

    // ignore first column of regression coefficients
    Matrix<double> Z(m_regressionCoefficients->GetRows()-1,m_regressionCoefficients->GetCols());
    for (unsigned int i=0;i<m_regressionCoefficients->GetCols();i++)
    {
        for (unsigned int j=0;j<m_regressionCoefficients->GetRows()-1;j++)
        {
            Z[j][i] = (*m_regressionCoefficients)[j+1][i];
        }
    }
    // transpose for warper
    Z = Transpose(Z);

    m_regressionCoefficients->ReSize(Z.GetRows(),Z.GetCols());

    for (unsigned int i=0;i<Z.GetRows();i++)
    {
        for (unsigned int j=0;j<Z.GetCols();j++)
        {
            (*m_regressionCoefficients)[i][j] = Z[i][j];
        }
    }

//    standardise();
}

void MMRNode::process()
{
    m_specimens = getInputPortData0();

    if (m_specimens == NULL || !m_specimens->isValid())
    {
        return;
    }

    Logger::getInstance()->log("[Regression Node] Processing started");


    if (listWidget->count() == 0)
    {
        FillLeftList();
    }

    if (listWidget_2->count() == 0)
    {
        FillRightList();
    }

    if (m_updated)
    {
        LoadData();
        Regress();

        setOutputPort0(m_regressionCoefficients);
        setOutputPort1(m_predictedValues);
        setOutputPort2(m_residuals);
        setOutputPort3(m_standardisedSpecimens);
        setOutputPort4(m_RSqr);
    }

    if (lineEdit->text() != "")
    {
        PrintOutput();
    }

    m_updated = false;

    Logger::getInstance()->log("[Regression Node] Processing finished");
}

void MMRNode::LoadData()
{
    if (radioButton->isChecked())   // shape variables
    {
        int dimensions = m_specimens->getLandmarkDimensions();
        int landmarks = m_specimens->getLandmarkCount();
        int individuals = m_specimens->getSize();
        m_Y.ReSize(individuals,dimensions*landmarks);

        for (int i=0;i<individuals;i++)
        {
            LandmarkSet *llm = m_specimens->getLandmarkSet(i);
            for (int j=0;j<landmarks;j++)
            {
                for (int k=0;k<dimensions;k++)
                {
                    m_Y[i][j*dimensions+k] = (*llm)[j][k];
                }
            }
        }
    }
    else if (radioButton_2->isChecked())   // form variables
    {
        int dimensions = m_specimens->getLandmarkDimensions();
        int landmarks = m_specimens->getLandmarkCount();
        int individuals = m_specimens->getSize();
        m_Y.ReSize(individuals,dimensions*landmarks+1);

        for (int i=0;i<individuals;i++)
        {
            LandmarkSet *llm = m_specimens->getLandmarkSet(i);
            for (int j=0;j<landmarks;j++)
            {
                for (int k=0;k<dimensions;k++)
                {
                    m_Y[i][j*dimensions+k] = (*llm)[j][k];
                }
            }
            m_Y[i][landmarks*dimensions] = log(m_specimens->getCentroidSize(i));
        }
    }
    else    // other variables
    {
        int individuals = m_specimens->getSize();
        int lcols = listWidget->selectedItems().count();

        if (individuals > 0 && lcols > 0)
        {
            m_Y.ReSize(individuals,lcols);

            // label values
            Labels *lbls = m_specimens->getSpecimenLabels(0);
            QStringList labelNames = lbls->keys();

            if (!labelNames.empty())
            {
                // first create a matrix of label values
                Matrix<double> labelMatrix(individuals,labelNames.size());

                QVector<QStringList> strvalues(m_specimens->getSpecimenLabels(0)->size());
                for (unsigned int i=0;i<m_specimens->getSize();i++)
                {
                    Labels *lbl = m_specimens->getSpecimenLabels(i);
                    QList<ILabelValue*> labelValues = lbl->values();

                    for (int j=0;j<labelValues.size();j++)
                    {
                        if (labelValues[j]->getType()==ILabelValue::STRING_VALUE)
                        {
                            StringLabelValue * val = (StringLabelValue*)(labelValues[j]);
                            if (!strvalues[j].contains(val->getValue()))
                            {
                                labelMatrix[i][j] = strvalues[j].size();
                                strvalues[j].push_back(val->getValue());
                            }
                            else
                            {
                                labelMatrix[i][j] = strvalues[j].indexOf(val->getValue());
                            }
                        }
                        else if (labelValues[j]->getType()==ILabelValue::SCALAR_VALUE)
                        {
                            ScalarLabelValue * val = (ScalarLabelValue*)(labelValues[j]);
                            labelMatrix[i][j] = val->getValue();
                        }
                    }
                }

                QList<QListWidgetItem *> leftSelected = listWidget->selectedItems();

                vector<int> leftIndexes;
                for (int i=0;i<leftSelected.count();i++)
                {
                    leftIndexes.push_back(listWidget->row(leftSelected.at(i)));
                }
                sort(leftIndexes.begin(),leftIndexes.end());

                for (unsigned int j=0;j<leftIndexes.size();j++)
                {
                    if (leftIndexes[j]==0)
                    {
                        m_Y.setColumnLabel(j,"Centroid sizes");
                        for (int i=0;i<individuals;i++)
                        {
                            m_Y[i][j] = m_specimens->getCentroidSize(i);
                        }
                    }
                    else if (leftIndexes[j]==1)
                    {
                        m_Y.setColumnLabel(j,"Ln of centroid sizes");
                        for (int i=0;i<individuals;i++)
                        {
                            m_Y[i][j] = log(m_specimens->getCentroidSize(i));
                        }
                    }
                    else if (leftIndexes[j] < (int)labelMatrix.GetCols()+2)
                    {
                        m_Y.setColumnLabel(j,labelNames[leftIndexes[j]-2]);
                        for (int i=0;i<individuals;i++)
                        {
                            m_Y[i][j] = labelMatrix[i][leftIndexes[j]-2];
                        }
                    }
                    else
                    {
                        m_Y.setColumnLabel(j,m_additionalVariables->getColumnLabel(leftIndexes[j]-2-labelMatrix.GetCols()));
                        for (int i=0;i<individuals;i++)
                        {
                            m_Y[i][j] = (*m_additionalVariables)[i][leftIndexes[j]-2-labelMatrix.GetCols()];
                        }
                    }
                }
            }
        }
    }

    if (radioButton_4->isChecked())   // shape variables
    {
        int dimensions = m_specimens->getLandmarkDimensions();
        int landmarks = m_specimens->getLandmarkCount();
        int individuals = m_specimens->getSize();

        m_X.ReSize(individuals,dimensions*landmarks);

        for (int i=0;i<individuals;i++)
        {
            LandmarkSet *rlm = m_specimens->getLandmarkSet(i);
            for (int j=0;j<landmarks;j++)
            {
                for (int k=0;k<dimensions;k++)
                {
                    m_X[i][j*dimensions+k] = (*rlm)[j][k];
                }
            }
        }
    }
    else if (radioButton_5->isChecked())   // form variables
    {
        int dimensions = m_specimens->getLandmarkDimensions();
        int landmarks = m_specimens->getLandmarkCount();
        int individuals = m_specimens->getSize();

        m_X.ReSize(individuals,dimensions*landmarks+1);

        for (int i=0;i<individuals;i++)
        {
            LandmarkSet *rlm = m_specimens->getLandmarkSet(i);
            for (int j=0;j<landmarks;j++)
            {
                for (int k=0;k<dimensions;k++)
                {
                    m_X[i][j*dimensions+k] = (*rlm)[j][k];
                }
            }
            m_X[i][landmarks*dimensions] = log(m_specimens->getCentroidSize(i));
        }
    }
    else    // other variables
    {
        int individuals = m_specimens->getSize();
        int rcols = listWidget_2->selectedItems().count();

        if (individuals > 0 && rcols > 0)
        {
            m_X.ReSize(individuals,rcols);

            // label values
            Labels *lbls = m_specimens->getSpecimenLabels(0);
            QStringList labelNames = lbls->keys();

            if (!labelNames.empty())
            {
                // first create a matrix of label values
                Matrix<double> labelMatrix(individuals,labelNames.size());
                int selectedIndex = comboBox->currentIndex();
                comboBox->clear();

                QVector<QStringList> strvalues(m_specimens->getSpecimenLabels(0)->size());
                for (unsigned int i=0;i<m_specimens->getSize();i++)
                {
                    Labels *lbl = m_specimens->getSpecimenLabels(i);
                    QList<ILabelValue*> labelValues = lbl->values();

                    for (int j=0;j<labelValues.size();j++)
                    {
                        if (labelValues[j]->getType()==ILabelValue::STRING_VALUE)
                        {
                            StringLabelValue * val = (StringLabelValue*)(labelValues[j]);
                            if (!strvalues[j].contains(val->getValue()))
                            {
                                labelMatrix[i][j] = strvalues[j].size();
                                strvalues[j].push_back(val->getValue());
                            }
                            else
                            {
                                labelMatrix[i][j] = strvalues[j].indexOf(val->getValue());
                            }
                        }
                        else if (labelValues[j]->getType()==ILabelValue::SCALAR_VALUE)
                        {
                            ScalarLabelValue * val = (ScalarLabelValue*)(labelValues[j]);
                            labelMatrix[i][j] = val->getValue();
                        }
                    }
                }

                QList<QListWidgetItem *> rightSelected = listWidget_2->selectedItems();

                vector<int> rightIndexes;
                for (int i=0;i<rightSelected.count();i++)
                {
                    rightIndexes.push_back(listWidget_2->row(rightSelected.at(i)));
                }
                sort(rightIndexes.begin(),rightIndexes.end());

                for (unsigned int j=0;j<rightIndexes.size();j++)
                {
                    if (rightIndexes[j]==0)
                    {
                        m_X.setColumnLabel(j,"Centroid sizes");
                        comboBox->addItem("Centroid sizes");
                        for (int i=0;i<individuals;i++)
                        {
                            m_X[i][j] = m_specimens->getCentroidSize(i);
                        }
                    }
                    else if (rightIndexes[j]==1)
                    {
                        m_X.setColumnLabel(j,"Ln of centroid sizes");
                        comboBox->addItem("Ln of centroid sizes");
                        for (int i=0;i<individuals;i++)
                        {
                            m_X[i][j] = log(m_specimens->getCentroidSize(i));
                        }
                    }
                    else if (rightIndexes[j] < (int)labelMatrix.GetCols()+2)
                    {
                        m_X.setColumnLabel(j,labelNames[rightIndexes[j]-2]);
                        comboBox->addItem(labelNames[rightIndexes[j]-2]);
                        for (int i=0;i<individuals;i++)
                        {
                            m_X[i][j] = labelMatrix[i][rightIndexes[j]-2];
                        }
                    }
                    else
                    {
                        m_X.setColumnLabel(j,m_additionalVariables->getColumnLabel(rightIndexes[j]-2-labelMatrix.GetCols()));
                        comboBox->addItem(m_additionalVariables->getColumnLabel(rightIndexes[j]-2-labelMatrix.GetCols()));
                        for (int i=0;i<individuals;i++)
                        {
                            m_X[i][j] = (*m_additionalVariables)[i][rightIndexes[j]-2-labelMatrix.GetCols()];
                        }
                    }
                }
                if (comboBox->count() > selectedIndex )
                {
                    comboBox->setCurrentIndex(selectedIndex);
                }
            }
        }
    }
}

void MMRNode::accept()
{
    if (m_updated)
        emit dataReady();
    QDialog::accept();
}

bool MMRNode::isUpdated() const
{
    return m_updated;
}

QString MMRNode::toString() const
{
    QString nodeData = "";
    if (radioButton->isChecked())
    {
        nodeData += "1";
    }
    else
    {
        nodeData += "0";
    }
    nodeData += "$$";
    if (radioButton_2->isChecked())
    {
        nodeData += "1";
    }
    else
    {
        nodeData += "0";
    }
    nodeData += "$$";
    if (radioButton_3->isChecked())
    {
        nodeData += "1";
    }
    else
    {
        nodeData += "0";
    }
    nodeData += "$$";
    if (radioButton_4->isChecked())
    {
        nodeData += "1";
    }
    else
    {
        nodeData += "0";
    }
    nodeData += "$$";
    if (radioButton_5->isChecked())
    {
        nodeData += "1";
    }
    else
    {
        nodeData += "0";
    }
    nodeData += "$$";
    if (radioButton_6->isChecked())
    {
        nodeData += "1";
    }
    else
    {
        nodeData += "0";
    }

    nodeData += "$$";
    nodeData += lineEdit->text();

    nodeData += "$$";
    nodeData += lineEdit_2->text();

    nodeData += "$$";
    QString idx;
    idx.setNum(comboBox->currentIndex());
    nodeData += idx;
    nodeData += "$$";

    Labels *lbl = m_specimens->getSpecimenLabels(0);
    QStringList labelNames = lbl->keys();

    nodeData += "Centroid sizes";
    nodeData += "|";
    nodeData += "Ln of centroid sizes";
    nodeData += "|";

    for (int i=0;i<labelNames.size();i++)
    {
        nodeData += labelNames[i];
        nodeData += "|";
    }

    if (m_additionalVariables != NULL && m_additionalVariables->isValid())
    {
        if (m_additionalVariables->GetRows() == m_specimens->getSize())
        {
            for (unsigned int i=0;i<m_additionalVariables->GetCols();i++)
            {
                nodeData += m_additionalVariables->getColumnLabel(i);
                nodeData += "|";
            }
        }
    }
    nodeData += "$$";
    for (int i=0;i<listWidget->count();i++)
    {
        if (listWidget->item(i)->isSelected())
        {
            nodeData += "1|";
        }
        else
        {
            nodeData += "0|";
        }
    }
    nodeData += "$$";
    for (int i=0;i<listWidget_2->count();i++)
    {
        if (listWidget_2->item(i)->isSelected())
        {
            nodeData += "1|";
        }
        else
        {
            nodeData += "0|";
        }
    }
    return  nodeData;
}

void MMRNode::fromString(const QString& params)
{
    QStringList values = params.split("$$");
    if (values.size() == 6)
    {
        radioButton->setChecked(values[0]=="1");
        radioButton_2->setChecked(values[1]=="1");
        radioButton_3->setChecked(values[2]=="1");
        radioButton_4->setChecked(values[3]=="1");
        radioButton_5->setChecked(values[4]=="1");
        radioButton_6->setChecked(values[5]=="1");
    }
    else if (values.size() > 6)
    {
        radioButton->setChecked(values[0]=="1");
        radioButton_2->setChecked(values[1]=="1");
        radioButton_3->setChecked(values[2]=="1");
        radioButton_4->setChecked(values[3]=="1");
        radioButton_5->setChecked(values[4]=="1");
        radioButton_6->setChecked(values[5]=="1");

        lineEdit->setText(values[6]);
        lineEdit_2->setText(values[7]);

        QStringList items = values[9].split("|");
        QStringList leftSelected = values[10].split("|");
        QStringList rightSelected = values[11].split("|");

        for (int i=0;i<leftSelected.count();i++)
        {
            listWidget->addItem(items[i]);
            listWidget_2->addItem(items[i]);
        }

        for (int i=0;i<listWidget->count();i++)
        {
            if (leftSelected[i] == "1")
                listWidget->setItemSelected ( listWidget->item(i), true );
        }

        for (int i=0;i<listWidget_2->count();i++)
        {
            if (rightSelected[i] == "1")
            {
                listWidget_2->setItemSelected ( listWidget_2->item(i), true );
                comboBox->addItem(items[i]);
            }
        }
        comboBox->setCurrentIndex(values[8].toInt());

        leftRadioChanged();
        rightRadioChanged();
    }
    m_updated = true;
}

bool MMRNode::inputChanged(int inputIndex)
{
    switch (inputIndex)
    {
    case 0: // Specimens
        FillLeftList();
        FillRightList();
        break;
    case 1: // Additional variables
        FillLeftList();
        FillRightList();
        break;
    default:
        break;
    }
    return true;
}

void MMRNode::FillRightList()
{
    m_specimens = getInputPortData0();
    if (m_specimens == NULL || !m_specimens->isValid())
    {
        return;
    }

    QVector<int> sel;
    for (int i=0;i<listWidget_2->count();i++)
    {
        if (listWidget_2->item(i)->isSelected())
            sel.push_back(i);
    }

    listWidget_2->clear();
    listWidget_2->addItem("Centroid sizes");
    listWidget_2->addItem("Ln of centroid sizes");

    Labels *lbl = m_specimens->getSpecimenLabels(0);
    QStringList labelNames = lbl->keys();

    for (int i=0;i<labelNames.size();i++)
    {
        listWidget_2->addItem(labelNames[i]);
    }

    m_additionalVariables = getInputPortData1();
    if (m_additionalVariables == NULL || !m_additionalVariables->isValid())
    {
        // do nothing
    }
    else
    {
        if (m_additionalVariables->GetRows() == m_specimens->getSize())
        {
            for (unsigned int i=0;i<m_additionalVariables->GetCols();i++)
            {
                listWidget_2->addItem(m_additionalVariables->getColumnLabel(i));
            }
        }
    }

    for (int i=0;i<sel.size();i++)
    {
        listWidget_2->setItemSelected ( listWidget_2->item(sel[i]), true );
    }
}

void MMRNode::FillLeftList()
{
    m_specimens = getInputPortData0();

    if (m_specimens == NULL || !m_specimens->isValid())
    {
        return;
    }

    QVector<int> sel;
    for (int i=0;i<listWidget->count();i++)
    {
        if (listWidget->item(i)->isSelected())
            sel.push_back(i);
    }

    listWidget->clear();
    listWidget->addItem("Centroid sizes");
    listWidget->addItem("Ln of centroid sizes");

    Labels *lbl = m_specimens->getSpecimenLabels(0);
    QStringList labelNames = lbl->keys();

    for (int i=0;i<labelNames.size();i++)
    {
        listWidget->addItem(labelNames[i]);
    }

    m_additionalVariables = getInputPortData1();

    if (m_additionalVariables == NULL || !m_additionalVariables->isValid())
    {
        // do nothing
    }
    else
    {
        if (m_additionalVariables->GetRows() == m_specimens->getSize())
        {
            for (unsigned int i=0;i<m_additionalVariables->GetCols();i++)
            {
                listWidget->addItem(m_additionalVariables->getColumnLabel(i));
            }
        }
    }
    for (int i=0;i<sel.size();i++)
    {
        listWidget->setItemSelected ( listWidget->item(sel[i]), true );
    }
}


void MMRNode::PrintOutput()
{
    QString fileName = lineEdit->text();
    QFile file(fileName);

    ofstream outfile(fileName.toStdString().c_str());

    if (outfile.fail())
    {
        Logger::getInstance()->log("[Regression Node] Error opening output file: " + fileName, Logger::RUN_ERROR);
        return;
    }

    //get current date
    QDate date = QDate::currentDate();
    QString dateString = date.toString();

    //get current time
    QTime time = QTime::currentTime();
    QString timeString = time.toString();

    //get current date and time
    QDateTime dateTime = QDateTime::currentDateTime();
    QString dateTimeString = dateTime.toString();

    outfile << "EVAN Toolbox" << endl;
    outfile << "Multiple Multivariate Regression" << endl;

    outfile << endl;

    outfile << "Date:\t" << dateString.toStdString() << endl;
    outfile << "Time:\t" << timeString.toStdString() << endl;

    outfile << endl;

    outfile << "Dependent variables" << endl;
    outfile << m_Y << endl;

    outfile << endl;

    outfile << "Independent variables" << endl;
    outfile << m_X << endl;

    outfile << endl;

    outfile << "R squared" << endl;
    outfile << (*m_RSqr) << endl;

    outfile << endl;

    Logger::getInstance()->log("[Regression Node] Output written to file: " + fileName);
}

void MMRNode::standardise()
{
    if (!m_mmr.IsCalculated())
    {
        return;
    }

    if (comboBox->count() > 0)
    {
        int landmarks = m_specimens->getLandmarkCount();
        int dimensions = m_specimens->getLandmarkDimensions();
        int index = comboBox->currentIndex();

        for (unsigned int i=0;i<m_specimens->getSize();i++)
        {
            LandmarkSet *lmks = m_specimens->getLandmarkSet(i);
            LandmarkSet *stand = m_standardisedSpecimens->getLandmarkSet(i);

            double x0 = 0.0;
            for (int j=0;j<landmarks;j++)
            {
                for (int k=0;k<dimensions;k++)
                {
                    double beta = (*m_regressionCoefficients)[index][j*dimensions+k];
                    (*stand)[j][k] = (*lmks)[j][k] - beta * (m_X[index][i] - x0);
                }
            }
        }

        setOutputPort3(m_standardisedSpecimens);
        emit dataReady();
    }
}

void MMRNode::comboChanged()
{
    lineEdit_2->setText("0.0");
    standardise();
}

void MMRNode::browseOutputFile()
{
    QString fileName = QFileDialog::getSaveFileName(this, tr("Select a file for Regression output"),
                       QString(),
                       tr("All files (*)\nText files (*.txt)"));

    if (fileName != "" && !fileName.isEmpty())
    {
        lineEdit->setText(fileName);
    }
}

