// wdkg 2008-2010

#ifndef EW_BBOX3_H_INCLUDED
#define EW_BBOX3_H_INCLUDED 1

/// 3D Bounding Box

/// EW_Bbox3 represents bounding boxes in @f$\mathbb{R}^3@f$.
///
/// EW_Bbox3 is a POD type class.
///
/// EW_Bbox3 is declared in @c ew/bbox3.h.

class EW_Bbox3 {
public:
  double min[3];
  double max[3];
  inline bool get_empty() const;
  double get_radius_center(double *c) const;
  inline void set_to_empty();
  template<typename T> inline void set_to_point(const T *pt);
  template<typename T> inline void set_to_points(const T *pt, int n);
  template<typename T> inline void add(const T *pt);
  template<typename T> inline void add(const T *pt, int n);
  void set_to_union(const EW_Bbox3 *b1, const EW_Bbox3 *b2);
  bool operator==(const EW_Bbox3 &a) const;
  inline bool operator!=(const EW_Bbox3 &a) const;
  static const EW_Bbox3 empty_box;
};

/// @var double EW_Bbox3::min[3]
/// The minima in the 3 coordinate directions.
/// Except for the empty box, these should all be finite and no greater than
/// the corresponding maxima.

/// @var double EW_Bbox3::max[3]
/// The maxima in the 3 coordinate directions.
/// Except for the empty box, these should all be finite and no less than the
/// corresponding minima.

/// @fn void EW_Bbox3::set_to_union(const EW_Bbox3 *b1, const EW_Bbox3 *b2)
/// This sets the bounding box to the union of the specified bounding boxes.
/// @param b1, b2
/// The boxes to combine.

/// @fn bool EW_Bbox3::get_empty() const
/// An empty box is a box with all minima infinity and all maxima negative
/// infinity.
/// @return
///   @c true if the box is the empty box.

/// @fn double EW_Bbox3::get_radius_center(double *c) const
/// This calculates the center and circumradius of a box.
/// @pre The box is not empty.
/// @param[out] c
///   This points to an array of size 3 where the center coordinates should
///   be stored.
/// @return
///   The radius.

/// @fn template<typename T> inline void EW_Bbox3::set_to_point(const T *pt)
/// This sets the box to be the bounding box of the point.
/// @pre T is either float or double and the points are all finite.
/// @param[in] pt
///   This points to an array of size 3 containing the coordinates of the
///   point.
///   They should be finite.

/// @fn template<typename T> inline void EW_Bbox3::set_to_points(const T *pt,
///  int n)
/// This sets the box to be the bounding box of the points.
/// @pre T is either float or double and the points are all finite.
/// @param[in] pt
///   This points to an array of size 3 * @a n containing the coordinates of
///   the points.
///   They should be finite.
/// @param[in] n
///   The number of points.

/// @fn template<typename T> inline void EW_Bbox3::add(const T *pt)
/// This extends the box, if necessary, to include the point.
/// @pre T is either float or double and the points are all finite.
/// @param[in] pt
///   This points to an array of size 3 containing the coordinates of the
///   point.
///   They should be finite.

/// @fn template<typename T> inline void EW_Bbox3::add(const T *pt, int n)
/// This extends the box, if necessary, to include points.
/// @pre T is either float or double and the points are all finite.
/// @param[in] pt
///   This points to an array of size 3 * @a n containing the coordinates of
///   the points.
/// @param[in] n
///   The number of points.

/// @fn bool EW_Bbox3::operator==(const EW_Bbox3 &a) const
/// This compares member by member.

/// @fn inline bool EW_Bbox3::operator!=(const EW_Bbox3 &a) const
/// This compares member by member.

/// @var const EW_Bbox3 EW_Bbox3::empty_box
/// This is an empty box.
/// All minima are infinity, all maxima are negative infinity.

// The rest of this file is private.

inline bool
EW_Bbox3::get_empty() const
{
  return (min[0] > max[0]);
}

inline void
EW_Bbox3::set_to_empty()
{
  *this = empty_box;
}

template<typename T>
inline void
EW_Bbox3::set_to_point(const T *pt)
{
  for (int i = 0; i < 3; i++) {
    min[i] = max[i] = static_cast<double>(pt[i]);
  }
}

template<typename T>
inline void
EW_Bbox3::set_to_points(const T *pt, int n)
{
  set_to_point(pt);
  add(pt + 3, n - 1);
}

template<typename T>
inline void
EW_Bbox3::add(const T *pt)
{
  for (int i = 0; i < 3; i++) {
    if (pt[i] < min[i]) {
      min[i] = static_cast<double>(pt[i]);
    }
    if (pt[i] > max[i]) {
      max[i] = static_cast<double>(pt[i]);
    }
  }
}

template<typename T>
inline void
EW_Bbox3::add(const T *pt, int n)
{
  for (int i = 0; i < n; i++) {
    add(pt + 3 * i);
  }
}

inline bool
EW_Bbox3::operator!=(const EW_Bbox3 &a) const
{
  return !operator==(a);
}

#endif
