// wdkg 2008-2010

#ifndef EW_DATAFLOW_CURVE3_H_INCLUDED
#define EW_DATAFLOW_CURVE3_H_INCLUDED 1

#include <memory>
#include "ew/bbox3.h"
#include "ew/dataflow_network.h"
#include "ew/dataflow_node.h"

class EWIndex3;

//XXX separate dlist for edges for use by warped curve
//XXX revert to generating dlist here as a update_ method (cf notes_nodes)
//XXX mechanism for optionally responding to file changes
//XXX   check for file modification when load, function for global reload
//XXX check dlist freed when no items using it

/// Curve Node Base Class

/// EW_Dataflow_Curve3 is the base class for nodes containing piecewise-linear
/// curves in @f$\mathbb{R}^3@f$.
///
/// EW_Dataflow_Curve3 is a class without assignment or comparison.
/// There are private member variables.
///
/// EW_Dataflow_Curve3 is declared in @c ew/dataflow_curve3.h.
///
/// For efficient rendering of the same curve in multiple windows, an OpenGL
/// display list is constructed containing the basic curve geometry commands.
/// All windows displaying a given curve must share display lists.

class EW_Dataflow_Curve3 : public EW_Dataflow_Node {
public:
  inline int get_num_points() const;
  inline int get_num_edges() const;
  inline const float *get_points() const;
  inline const int *get_edges() const;
  inline const EW_Bbox3 *get_bbox() const;
  inline void make_index() const;
  inline bool index_is_made() const;
  double project(int *edge, double *coeffs, double *tangent, double *proj,
   double *inp) const;
protected:
  EW_Dataflow_Curve3(EW_Dataflow_Network *nw, int ndeps, const char *n);
  ~EW_Dataflow_Curve3();
  mutable int OutpNumPoints;
  mutable int OutpNumEdges;
  mutable const float *OutpPoints;
  mutable const int *OutpEdges;
  void reset_tangents() const;
  void reset_index() const;
protected:
//XXX make these private (needed for reset)
  mutable EW_Dataflow_Network::cycle_t update_cycle_index;
  mutable EW_Dataflow_Network::cycle_t update_cycle_tangents;
private:
  friend class EW_View3_Curve;
// EW_View3_Curve
  inline unsigned int get_dlist() const;
  inline void set_hold_dlist(unsigned int dl) const;
  inline void hold_dlist() const;
  inline void release_dlist() const;
  inline EW_Dataflow_Network::cycle_t get_dlist_cycle() const;
  inline void set_dlist_cycle(EW_Dataflow_Network::cycle_t cycle) const;
// private
  void update_bbox() const;
  inline const float *get_tangents() const;
  void update_tangents() const;
  void update_index() const;
  mutable EW_Bbox3 OutpBbox;
  mutable std::vector<float> OutpTangents;
  const std::auto_ptr<EWIndex3> OutpIndex;
  mutable EW_Dataflow_Network::cycle_t update_cycle_bbox;
  mutable unsigned int dlist;
  mutable EW_Dataflow_Network::cycle_t dlist_cycle;
  mutable int dlist_refs;
  const bool DbgOnSur;
};

// index is a pointer just to avoid having to define EWIndex3 in the
// exported header files.

/// @fn int EW_Dataflow_Curve3::get_num_points() const
/// This will force the network into an updating phase.
/// @return
///   The number of points used in the curve.

/// @fn int EW_Dataflow_Curve3::get_num_edges() const
/// This will force the network into an updating phase.
/// @return
///   The number of triangular edges in the curve.

/// @fn const float *EW_Dataflow_Curve3::get_points() const
/// This will force the network into an updating phase.
/// The pointer is valid until the node or a dependency of the node is changed.
/// @return
///   A pointer to the coordinates of the points used in the curve.

/// @fn const int *EW_Dataflow_Curve3::get_edges() const
/// This will force the network into an updating phase.
/// The pointer is valid until the node or a dependency of the node is changed.
/// @return
///   A pointer to the indices of the points used in the curve.

/// @fn const EW_Bbox3 *EW_Dataflow_Curve3::get_bbox() const
/// The pointer is valid until the node or a dependency of the node is changed.
/// @return
///   A pointer to the bounding box of the curve.

/// @fn inline void EW_Dataflow_Curve3::make_index() const
/// This makes the curve spatial index if it has not already been made since
/// the curve was last changed.
/// The spatial index is used by #project.

/// @fn inline bool EW_Dataflow_Curve3::index_is_made() const;
/// @return
///   @c true if the curve spatial index is up to date.

/// @fn double EW_Dataflow_Curve3::project(int *edge, double *coeffs,
///  double *tangent, double *proj, double *inp) const
/// This finds the nearest point on the curve to a given point.
/// The curve spatial index will be made if it has not already been made.
/// @param[out] edge
///   The index of the edge of the curve containing the nearest point.
/// @param[out] coeffs
///   The coefficients of the nearest point when expressed as a linear
///    combination of the vertices of the edge.
/// @param[out] tangent
///   The interpolated tangent at the nearest point.
///   If a sensible tangent cannot be calculated, an arbitrary unit vector
///   is returned.
/// @param[out] proj
///   The coordinates of the nearest point on the curve.
/// @param[in] inp
///   The coordinates of the original point.
/// @return
///   The distance from the original point to the projected point.

// The rest of this file is private.

inline int
EW_Dataflow_Curve3::get_num_points() const
{
  EW_Dataflow_Node::update();
  return OutpNumPoints;
}

inline int
EW_Dataflow_Curve3::get_num_edges() const
{
  EW_Dataflow_Node::update();
  return OutpNumEdges;
}

inline const float *
EW_Dataflow_Curve3::get_points() const
{
  EW_Dataflow_Node::update();
  return OutpPoints;
}

inline const int *
EW_Dataflow_Curve3::get_edges() const
{
  EW_Dataflow_Node::update();
  return OutpEdges;
}

inline const EW_Bbox3 *
EW_Dataflow_Curve3::get_bbox() const
{
  EW_Dataflow_Network::cycle_t v = EW_Dataflow_Node::get_version();
  if (update_cycle_bbox < v) {
    update_bbox();
    update_cycle_bbox = EW_Dataflow_Node::network->get_cycle();
  }
  return &OutpBbox;
}

inline const float *
EW_Dataflow_Curve3::get_tangents() const
{
  EW_Dataflow_Network::cycle_t v = EW_Dataflow_Node::get_version();
  if (update_cycle_tangents < v) {
    update_tangents();
    update_cycle_tangents = EW_Dataflow_Node::network->get_cycle();
  }
  return OutpTangents.empty() ? 0 : &(OutpTangents[0]);
}

inline void
EW_Dataflow_Curve3::make_index() const
{
  EW_Dataflow_Network::cycle_t v = EW_Dataflow_Node::get_version();
  if (update_cycle_index < v) {
    update_index();
    update_cycle_index = EW_Dataflow_Node::network->get_cycle();
  }
}

inline bool
EW_Dataflow_Curve3::index_is_made() const
{
  return (update_cycle_index >= EW_Dataflow_Node::get_version());
}

inline unsigned int
EW_Dataflow_Curve3::get_dlist() const
{
  return dlist;
}

inline void
EW_Dataflow_Curve3::set_hold_dlist(unsigned int dl) const
{
  dlist = dl;
  dlist_cycle = 0;
  dlist_refs++;
}

inline EW_Dataflow_Network::cycle_t
EW_Dataflow_Curve3::get_dlist_cycle() const
{
  return dlist_cycle;
}

inline void
EW_Dataflow_Curve3::set_dlist_cycle(EW_Dataflow_Network::cycle_t cycle) const
{
  dlist_cycle = cycle;
}

inline void
EW_Dataflow_Curve3::hold_dlist() const
{
  dlist_refs++;
}


inline void
EW_Dataflow_Curve3::release_dlist() const
{
  dlist_refs--;
  if (dlist_refs == 0) {
    dlist = 0;
    dlist_cycle = 0;
  }
}

#endif
