// wdkg 2008-2010

#ifndef EW_DATAFLOW_FORM3_H_INCLUDED
#define EW_DATAFLOW_FORM3_H_INCLUDED 1

#include "ew/bbox3.h"
#include "ew/dataflow_network.h"
#include "ew/dataflow_node.h"
#include "ew/form3.h"
#include "ew/form3_curve.h"
#include "ew/form3_embedding.h"
#include "ew/form3_pointset.h"
#include "ew/form3_reflection.h"
#include "ew/form3_surface.h"
#include "ew/form3_volume.h"

/// 3D Form Node

/// EW_Dataflow_Form3 is a node that manages a EW_Form3 instance.
///
/// EW_Dataflow_Form3 is a class without assignment or comparison.
/// There are private member variables.
///
/// EW_Dataflow_Form3 is declared in @c ew/dataflow_form3.h.
///
/// Nodes of this class do not depend on other nodes, but other nodes can
/// depend on them.
/// They are always considered valid.
///
/// Initially and when reset, the node is empty.

class EW_Dataflow_Form3 : public EW_Dataflow_Node {
public:
  explicit EW_Dataflow_Form3(EW_Dataflow_Network *nw);
  virtual void reset();
  void set_data(const EW_Form3 *data);
  inline const EW_Form3 *get_data() const;
  int set_volume(bool *replaced, const EW_Form3_Volume *vl);
  void remove_volume(int n);
  int set_surface(bool *replaced, const EW_Form3_Surface *sr);
  void remove_surface(int n);
  int set_curve(bool *replaced, const EW_Form3_Curve *cr);
  void remove_curve(int n);
  int set_pointset(bool *replaced, const EW_Form3_PointSet *ps);
  void set_pointset_location(int n, int i, const double *loc);
  void set_pointset_relax(int n, int i, int rdim, const double *rparam);
  void remove_pointset(int n);
  int set_embedding(bool *replaced, const EW_Form3_Embedding *em);
  void remove_embedding(int n);
  int set_reflection(bool *replaced, const EW_Form3_Reflection *rf);
  void remove_reflection(int n);
  inline EW_Dataflow_Network::cycle_t get_change_cycle_association() const;
  inline EW_Dataflow_Network::cycle_t get_change_cycle_coords() const;
  inline EW_Dataflow_Network::cycle_t get_change_cycle_relax() const;
  inline const EW_Bbox3 *get_bbox() const;
private:
  virtual void update_helper() const;
  void update_bbox() const;
  EW_Form3 InpData;
  EW_Dataflow_Network::cycle_t change_cycle_coords;
  EW_Dataflow_Network::cycle_t change_cycle_relax;
  EW_Dataflow_Network::cycle_t change_cycle_association;
  mutable EW_Bbox3 OutpBbox;
  mutable EW_Dataflow_Network::cycle_t update_cycle_bbox;
  const bool DbgOn;
};

/// @fn explicit EW_Dataflow_Form3::EW_Dataflow_Form3(EW_Dataflow_Network *nw)
/// This creates a form3 node.
/// @param nw
///   The network this node should belong to.

/// @fn void EW_Dataflow_Form3::set_data(const EW_Form3 *data)
/// This sets the contents of the node.
/// @param data
///   A pointer to the EW_Form3 data to copy.

/// @fn inline const EW_Form3 *EW_Dataflow_Form3::get_data() const
/// @return
///   A pointer to the current EW_Form3 data contained in the node.

/// @fn int EW_Dataflow_Form3::set_volume(bool *replaced,
///  const EW_Form3_Volume *vl)
/// This adds or replaces a volume in the EW_Form3 data managed by the node.
/// @param[out] replaced
///   @c true if the volume replaced an existing volume.
/// @param vl
///   A pointer to the volume data to copy.
/// @return
///   The index of the new volume in EW_Form3::volumes.

/// @fn void EW_Dataflow_Form3::remove_volume(int n)
/// This deletes a volume in the EW_Form3 managed by the node.
/// @param n
///   The index of the volume to delete.

/// @fn int EW_Dataflow_Form3::set_surface(bool *replaced,
///  const EW_Form3_Surface *sr)
/// This adds or replaces a surface in the EW_Form3 data managed by the node.
/// @param[out] replaced
///   @c true if the surface replaced an existing surface.
/// @param sr
///   A pointer to the surface data to copy.
/// @return
///   The index of the new surface in EW_Form3::surfaces.

/// @fn void EW_Dataflow_Form3::remove_surface(int n)
/// This deletes a surface in the EW_Form3 managed by the node.
/// @param n
///   The index of the surface to delete.

/// @fn int EW_Dataflow_Form3::set_curve(bool *replaced,
///  const EW_Form3_Curve *cr)
/// This adds or replaces a curve in the EW_Form3 data managed by the node.
/// @param[out] replaced
///   @c true if the curve replaced an existing curve.
/// @param cr
///   A pointer to the curve data to copy.
/// @return
///   The index of the new curve in EW_Form3::curves.

/// @fn void EW_Dataflow_Form3::remove_curve(int n)
/// This deletes a curve in the EW_Form3 managed by the node.
/// @param n
///   The index of the curve to delete.

/// @fn int EW_Dataflow_Form3::set_pointset(bool *replaced,
///  const EW_Form3_PointSet *ps)
/// This adds or replaces a point set in the EW_Form3 data managed by the node.
/// @param[out] replaced
///   @c true if the point set replaced an existing point set.
/// @param ps
///   A pointer to the point set data to copy.
/// @return
///   The index of the new or replaced pointset.

/// @fn void EW_Dataflow_Form3::set_pointset_location(int n, int i,
///  const double *loc);
/// This changes the coordinates of an element of the point set.
/// @param n
///   The index of the point set.
/// @param i
///   The index of the element in the point set.
/// @param[in] loc
///   A pointer to the new coordinates of this entry.

/// @fn void EW_Dataflow_Form3::set_pointset_relax(int n, int i, int rdim,
///  const double *rparam);
/// This changes the relaxation parameters of an element of the point set.
/// @param n
///   The index of the point set.
/// @param i
///   The index of the element in the point set.
/// @param[in] rdim
///   The new relax_dims of this entry.
/// @param[in] rparam
///   A pointer to the new relax_params of this entry.

/// @fn void EW_Dataflow_Form3::remove_pointset(int n)
/// This deletes a point set in the EW_Form3 managed by the node.
/// @param n
///   The index of the point set to delete.

/// @fn int EW_Dataflow_Form3::set_embedding(bool *replaced,
///  const EW_Form3_Embedding *em)
/// This adds or replaces a embedding in the EW_Form3 data managed by the node.
/// @param[out] replaced
///   @c true if the embedding replaced an existing embedding.
/// @param em
///   A pointer to the embedding data to copy.
/// @return
///   The index of the new embedding in EW_Form3::embeddings.

/// @fn void EW_Dataflow_Form3::remove_embedding(int n)
/// This deletes a embedding in the EW_Form3 managed by the node.
/// @param n
///   The index of the embedding to delete.

/// @fn int EW_Dataflow_Form3::set_reflection(bool *replaced,
///  const EW_Form3_Reflection *rf)
/// This adds or replaces a reflection in the EW_Form3 data managed by the
/// node.
/// @param[out] replaced
///   @c true if the reflection replaced an existing reflection.
/// @param rf
///   A pointer to the reflection data to copy.
/// @return
///   The index of the new reflection in EW_Form3::reflections.

/// @fn void EW_Dataflow_Form3::remove_reflection(int n)
/// This deletes a reflection in the EW_Form3 managed by the node.
/// @param n
///   The index of the reflection to delete.

// Doxygen comment without long line or g++ multiline comment warning.
/*!
 * @fn inline EW_Dataflow_Network::cycle_t \
 *  EW_Dataflow_Form3::get_change_cycle_association() const
 * @return
 *   The last cycle the form was changed in a way that effects landmark matching.
 */

// Doxygen comment without long line or g++ multiline comment warning.
/*!
 * @fn inline EW_Dataflow_Network::cycle_t \
 *  EW_Dataflow_Form3::get_change_cycle_coords() const
 * @return
 *   The last cycle that any location of a landmark was changed.
 */

// Doxygen comment without long line or g++ multiline comment warning.
/*!
 * @fn inline EW_Dataflow_Network::cycle_t \
 *  EW_Dataflow_Form3::get_change_cycle_relax() const
 * @return
 *   The last cycle that any relaxation of a landmark was changed.
 */

/// @fn const EW_Bbox3 *EW_Dataflow_Form3::get_bbox() const
/// The pointer until the node or a dependency of the node is changed.
/// @return
///   A pointer to the bounding box of the points in the form.

// The rest of this file is private.

inline const EW_Form3 *
EW_Dataflow_Form3::get_data() const
{
  return &InpData;
}

inline EW_Dataflow_Network::cycle_t
EW_Dataflow_Form3::get_change_cycle_association() const
{
  return change_cycle_association;
}

inline EW_Dataflow_Network::cycle_t
EW_Dataflow_Form3::get_change_cycle_coords() const
{
  return change_cycle_coords;
}

inline EW_Dataflow_Network::cycle_t
EW_Dataflow_Form3::get_change_cycle_relax() const
{
  return change_cycle_relax;
}

inline const EW_Bbox3 *
EW_Dataflow_Form3::get_bbox() const
{
  EW_Dataflow_Network::cycle_t v = EW_Dataflow_Node::get_version();
  if (update_cycle_bbox < v) {
    update_bbox();
    update_cycle_bbox = EW_Dataflow_Node::network->get_cycle();
  }
  return &OutpBbox;
}

#endif
