// wdkg 2008-2010

#ifndef EW_DATAFLOW_NETWORK_H_INCLUDED
#define EW_DATAFLOW_NETWORK_H_INCLUDED 1

#include <string>
#include <vector>
#include "ew/dataflow_notifier_state.h"
#include "ew/dig3_state.h"
#include "ew/view3_state.h"

class EW_Dataflow_Node;
class EW_Dataflow_Surface3_Explicit;
class EW_Dataflow_Curve3_Explicit;
class EWDebug;

/// Lightweight Dataflow Network
// XXX method to free unreferenced cache elements
// XXX  possibly limit number of cached surfaces
// XXX support reloading of changed files in cache

/// EW_Dataflow_Network contains the common data of a lightweight dataflow
/// network.
///
/// EW_Dataflow_Network does not support assignment or comparison.
/// It contains private members and undocumented members.
///
/// EW_Dataflow_Network is declared in @c ew/dataflow_network.h.
///
/// To keep track of what calculations in the network are up to date, and which
/// need to be redone, EW_Dataflow_Network keeps a count of the cycle.
/// A cycle consists of a an initial phase, when changes are made to
/// nodes of the network, followed by a second phase, when updates of various
/// calculations are made.
/// Requesting updated output of a node puts the network into the update phase
/// of the same cycle, if it is not already in an update phase.
/// Changing a node puts the network into the change phase of the next cycle,
/// if it is not already in a change phase.
/// Initially the network is in the change phase of cycle 1.

class EW_Dataflow_Network {
public:
  typedef unsigned int cycle_t;
  EW_Dataflow_Network();
  ~EW_Dataflow_Network();
  inline EW_Dataflow_Network::cycle_t get_cycle() const;
  const EW_Dataflow_Surface3_Explicit *cached_surface(const char *filename);
  const EW_Dataflow_Curve3_Explicit *cached_curve(const char *filename);
//XXX  void uncache_surface(const EW_Dataflow_Surface3_Explicit *sur);
// private to library
// accessible only by EW_Dataflow_Notifier
  EW_Dataflow_Notifier_State NotifierState;
// accessible only by EW_View3_Widget
  EW_View3_State View3State;
// accessible only by EW_Dig3
  EW_Dig3_State Dig3State;
private:
  friend class EW_Dataflow_Node;
// EW_Dataflow_Node
  inline void start_changing();
  inline void start_updating();
// private
  void start_changing_helper();
  void start_updating_helper();
  EW_Dataflow_Network::cycle_t cycle;
  bool IsUpdating;
  std::vector<EW_Dataflow_Surface3_Explicit *> cached_surfaces;
  std::vector<EW_Dataflow_Curve3_Explicit *> cached_curves;
// EW_Dataflow_Node
  unsigned int DbgLastNodeN;
// private
  const EWDebug &Dbg;
  const bool DbgOn;
public:
// private to library
  const int DbgN;
private:
  const std::string DbgStr;
public:
// private to library
  const char *const DbgId;
private:
// undefined
  EW_Dataflow_Network(const EW_Dataflow_Network &rhs);
  EW_Dataflow_Network &operator=(const EW_Dataflow_Network &rhs);
};

/// @typedef unsigned long EW_Dataflow_Network::cycle_t
/// This is used for cycle counters.

// It could be unsigned long long it there was any likelihood of overflow.
// Then we'd need to adjust the format specifiers in EWDebug::print's.

/// @fn EW_Dataflow_Network::EW_Dataflow_Network()
/// This creates an empty network.

/// @fn EW_Dataflow_Network::~EW_Dataflow_Network()
/// This destroys the network.
/// All nodes and other objects using this network must already have been
/// destroyed.

/// @fn EW_Dataflow_Network::cycle_t EW_Dataflow_Network::get_cycle() const
/// @return
///   The current cycle number.

/// @fn EW_Dataflow_Curve3_Explicit *EW_Dataflow_Network::cached_curve(
///  const char *filename)
/// The first time this is called with a given filename, it reads the curve
/// from the file, creates a new explicit curve node using it, and returns
/// a pointer to it.
/// Subsequently it increments the same node's reference count and returns a
/// pointer to it.
/// When the pointer to the node is no longer needed, the reference should be
/// released with EW_Dataflow_Node::decr_ref_count.
/// The cache retains a reference to the node, so it will not be deleted until
/// all other references are released and the network is destroyed.
/// @param filename
///   The file to read.
/// @exception std::bad_alloc
/// @exception EW_Curve3_IOError

/// @fn EW_Dataflow_Surface3_Explicit *EW_Dataflow_Network::cached_surface(
///  const char *filename)
/// The first time this is called with a given filename, it reads the surface
/// from the file, creates a new explicit surface node using it, and returns
/// a pointer to it.
/// Subsequently it increments the same node's reference count and returns a
/// pointer to it.
/// When the pointer to the node is no longer needed, the reference should be
/// released with EW_Dataflow_Node::decr_ref_count.
/// The cache retains a reference to the node, so it will not be deleted until
/// all other references are released and the network is destroyed.
/// @param filename
///   The file to read.
/// @exception std::bad_alloc
/// @exception EW_Surface3_IOError

// The rest of this file is private.

inline EW_Dataflow_Network::cycle_t
EW_Dataflow_Network::get_cycle() const
{
  return cycle;
}

inline void
EW_Dataflow_Network::start_changing()
{
  if (IsUpdating || !NotifierState.list_is_empty()) {
    start_changing_helper();
  }
}

inline void
EW_Dataflow_Network::start_updating()
{
  if (!IsUpdating) {
    start_updating_helper();
  }
}

#endif
