// wdkg 2008-2010

#ifndef EW_DATAFLOW_NODE_H_INCLUDED
#define EW_DATAFLOW_NODE_H_INCLUDED 1

#include <stdexcept>
#include <string>
#include <vector>
#include "ew/dataflow_network.h"

class EWDebug;

//XXX abstract EW_Dataflow_Points, put get_bbox here as well as warping
// stuff.

/// Node Base Class

/// EW_Dataflow_Node is the base class for nodes in a EW_Dataflow_Network.
///
/// EW_Dataflow_Node does not support assignment or comparison.
/// It contains private members, protected members and undocumented members.
///
/// EW_Dataflow_Node is declared in @c ew/dataflow_node.h.
///
/// Nodes are referenced counted and can only be deleted indirectly by
/// calling #decr_ref_count.
/// The reference count is set to 1 by the constructor.
///
/// Depending on its parameters, a node is considered valid or invalid.
/// What constitutes a valid state depends on the specific node class.
///
/// A node can depend on other nodes.
/// The possible dependencies are defined by the specific node class.
///
/// After the EW_Dataflow_Network has been destroyed, the only
/// EW_Dataflow_Method that may be called is decr_ref_count.

// ndeps must be at most INT_MAX.

class EW_Dataflow_Node {
public:
  virtual void reset() = 0;
  void incr_ref_count() const;
  void decr_ref_count() const;
  inline EW_Dataflow_Network::cycle_t get_change_cycle() const;
  inline EW_Dataflow_Network::cycle_t get_version() const;
  inline bool get_valid() const;
  EW_Dataflow_Network *const network;
protected:
  EW_Dataflow_Node(EW_Dataflow_Network *nw, int ndeps,
   const char *dbg_final_class_name);
  virtual ~EW_Dataflow_Node();
  void set_dependency(int i, const char *dbg_desc,
   const EW_Dataflow_Node *dep);
  inline void update() const;
  inline void record_change();
  EW_Dataflow_Network::cycle_t change_cycle;
  std::vector<const EW_Dataflow_Node *> InpDependencies;
  std::vector<EW_Dataflow_Network::cycle_t> change_cycle_dependencies;
  mutable bool OutpValid;
  mutable EW_Dataflow_Network::cycle_t OutpVersion;
  mutable EW_Dataflow_Network::cycle_t update_cycle;
private:
  virtual void update_helper() const = 0;
  bool depends_on(EW_Dataflow_Node *node) const;
  mutable int ref_count;
protected:
  const EWDebug &Dbg;
private:
  const bool DbgOnNode;
  const int DbgN;
  std::string DbgStr;
public:
// private to library
  const char *const DbgId;
private:
// undefined
  EW_Dataflow_Node(const EW_Dataflow_Node &rhs);
  EW_Dataflow_Node &operator=(const EW_Dataflow_Node &rhs);
};

/// @fn void EW_Dataflow_Node::reset()
/// This resets a node to the state it had immediately after construction.

/// @fn void EW_Dataflow_Node::incr_ref_count() const
/// This increases the reference count.

/// @fn void EW_Dataflow_Node::decr_ref_count() const
/// This decreases the reference count.
/// If the reference count becomes zero, the node is deleted.

/// @fn EW_Dataflow_Network::cycle_t EW_Dataflow_Node::get_change_cycle() const
/// @return
///   @c The last cycle on which a change to the node was made.
//
/// @fn EW_Dataflow_Network::cycle_t EW_Dataflow_Node::get_version() const
/// The version of a node is the greater of the last network cycle that any
/// parameter of the node was changed, and the versions of any nodes that the
/// node depends on.
/// This will force the network into an updating phase.
/// @return
///   @c The version of the node.

/// @fn bool EW_Dataflow_Node::get_valid() const
/// This will force the network into an updating phase.
/// @return
///   @c true if the node is in a valid state.

/// @var EW_Dataflow_Network *const EW_Dataflow_Node::network
/// This points to the EW_Dataflow_Network the node belongs to.

// The rest of this file is private.

// Concrete node classes must implement Ctor, reset and update_helper.

// Any exception thrown by idle callbacks will be passed through here.
// It is permissible, and recommended, to call this before changes are made.
// That way if it throws an exception the network is left in a consistent
// state.

inline void
EW_Dataflow_Node::record_change()
{
  network->start_changing();
  change_cycle = network->get_cycle();
}

inline void
EW_Dataflow_Node::update() const
{
  if (update_cycle != network->get_cycle()) {
    network->start_updating();
    update_helper();
    update_cycle = network->get_cycle();
  }
}

inline EW_Dataflow_Network::cycle_t
EW_Dataflow_Node::get_change_cycle() const
{
  return change_cycle;
}

inline EW_Dataflow_Network::cycle_t
EW_Dataflow_Node::get_version() const
{
  update();
  return OutpVersion;
}

inline bool
EW_Dataflow_Node::get_valid() const
{
  update();
  return OutpValid;
}

#endif
