// wdkg 2008-2010

#ifndef EW_DATAFLOW_SPLINE3_H_INCLUDED
#define EW_DATAFLOW_SPLINE3_H_INCLUDED 1

#include <vector>
#include "ew/dataflow_form3.h"
#include "ew/dataflow_network.h"
#include "ew/dataflow_node.h"

/// 3D Spline Node

/// EW_Dataflow_Spline3 is a node that manages a thin-plate spline.
///
/// EW_Dataflow_Spline3 is a class without assignment or comparison.
/// There are private member variables.
///
/// EW_Dataflow_Spline3 is declared in @c ew/dataflow_spline3.h.
///
/// Nodes of this class depend on 2 nodes, the template and specimen
/// EW_Dataflow_Form3 nodes.
/// Nodes representing warped geometric objects will depend on a
/// EW_Dataflow_Spline3 node.
/// A EW_Dataflow_Spline3 is considered valid if it's template
/// and specimen dependencies have been set and if there are 4 or more
/// associated landmarks.
///
/// Initially and when reset, the node has neither template or specimen set.

class EW_Dataflow_Spline3 : public EW_Dataflow_Node {
public:
  explicit EW_Dataflow_Spline3(EW_Dataflow_Network *nw);
  virtual void reset();
  void set_template(const EW_Dataflow_Form3 *f);
  inline const EW_Dataflow_Form3 *get_template() const;
  inline EW_Dataflow_Network::cycle_t get_change_cycle_template() const;
  void set_specimen(const EW_Dataflow_Form3 *f);
  inline const EW_Dataflow_Form3 *get_specimen() const;
  inline EW_Dataflow_Network::cycle_t get_change_cycle_specimen() const;
  inline int get_n_lmks() const;
  inline int get_f_size() const;
  inline int lmk_index(int side, int point_set, int i) const;
  inline int lmk_pointset(int side, int lmk_index) const;
  inline int lmk_pointset_i(int side, int lmk_index) const;
  inline bool get_nonsingular() const;
  inline double get_energy() const;
  inline const double *get_optimized_lmk_images() const;
  void warp_points(double *outp, const double *inp, int l) const;
protected:
  mutable EW_Dataflow_Network::cycle_t update_cycle_association;
  mutable EW_Dataflow_Network::cycle_t update_cycle_interaction;
  mutable EW_Dataflow_Network::cycle_t update_cycle_factorization;
  mutable EW_Dataflow_Network::cycle_t update_cycle_spline;
private:
  virtual void update_helper() const;
  inline EW_Dataflow_Network::cycle_t get_version_association() const;
  inline EW_Dataflow_Network::cycle_t get_version_interaction() const;
  inline EW_Dataflow_Network::cycle_t get_version_factorization() const;
  inline EW_Dataflow_Network::cycle_t get_version_spline() const;
  mutable EW_Dataflow_Network::cycle_t OutpVersionAssociation;
  mutable EW_Dataflow_Network::cycle_t OutpVersionInteraction;
  mutable EW_Dataflow_Network::cycle_t OutpVersionFactorization;
  mutable EW_Dataflow_Network::cycle_t OutpVersionSpline;
// a
  inline void make_association() const;
  void update_association() const;
  void reset_association() const;
  mutable int OutpNLmks;
  mutable std::vector<int> OutpSplineToPs[2];
  mutable std::vector<int> OutpSplineToPsI[2];
  mutable std::vector<int> OutpPsToSpline[2];
// b
  inline void make_interaction() const;
  void update_interaction() const;
  void reset_interaction() const;
  mutable std::vector<double> OutpLmks;
  mutable std::vector<double> OutpInteraction;
// c
  inline void make_factorization() const;
  void update_factorization() const;
  void reset_factorization() const;
  mutable int OutpFSize;
  mutable bool OutpIsMixed;
  mutable bool OutpIsReduced;
  mutable bool OutpNonsingular;
  mutable double OutpNonsingularity;
  mutable std::vector<int> OutpRelaxDims;
  mutable std::vector<double> OutpRelaxParams;
  mutable std::vector<double> OutpFactorization;
  mutable std::vector<int> OutpPivots;
// d
  inline void make_spline() const;
  void update_spline() const;
  void reset_spline() const;
  mutable double OutpEnergy;
  mutable std::vector<double> OutpLmkImages;
  mutable std::vector<double> OutpSpline;
  mutable std::vector<double> OutpRelaxLmkImages;
  const bool DbgOn;
};

/// @fn void EW_Dataflow_Spline3::set_template(const EW_Dataflow_Form3 *f);
/// @param f
///   A pointer to the new template node.

/// @fn inline EW_Dataflow_Form3 *EW_Dataflow_Spline3::get_template() const;
/// @return
///   A pointer to the template node.

// Doxygen comment without long line or g++ multiline comment warning.
/*!
 * @fn inline EW_Dataflow_Network::cycle_t \
 *  EW_Dataflow_Spline3::get_change_cycle_template() const
 * @return
 *   The last cycle the template pointer was changed.
 */

/// @fn void EW_Dataflow_Spline3::set_specimen(const EW_Dataflow_Form3 *f);
/// @param f
///   A pointer to the new pecimen node.

/// @fn inline EW_Dataflow_Form3 *EW_Dataflow_Spline3::get_specimen() const;
/// @return
///   A pointer to the specimen node.

// Doxygen comment without long line or g++ multiline comment warning.
/*!
 * @fn inline EW_Dataflow_Network::cycle_t \
 *  EW_Dataflow_Spline3::get_change_cycle_specimen() const
 * @return
 *   The last cycle the specimen pointer was changed.
 */

//XXX export/changelog
// Doxygen comment without long line or g++ multiline comment warning.
/*
 * @fn inline EW_Dataflow_Network::cycle_t \
 * EW_Dataflow_Spline3::get_version_association() const
 * @return
 *   The last cycle the specimen pointer was changed.
 */

// Doxygen comment without long line or g++ multiline comment warning.
/*
 * @fn inline EW_Dataflow_Network::cycle_t \
 * EW_Dataflow_Spline3::get_version_interaction() const
 * @return
 *   The last cycle the specimen pointer was changed.
 */

// Doxygen comment without long line or g++ multiline comment warning.
/*
 * @fn inline EW_Dataflow_Network::cycle_t \
 * EW_Dataflow_Spline3::get_version_factorization() const
 * @return
 *   The last cycle the specimen pointer was changed.
 */

// Doxygen comment without long line or g++ multiline comment warning.
/*
 * @fn inline EW_Dataflow_Network::cycle_t \
 * EW_Dataflow_Spline3::get_version_spline() const
 * @return
 *   The last cycle the specimen pointer was changed.
 */

/// @fn inline int EW_Dataflow_Spline3::get_n_lmks() const
/// @return
///   The number of landmarks and semi-landmarks matched between template and
///   specimen.

/// @fn inline int EW_Dataflow_Spline3::get_f_size() const
/// @return
///   The algebraic dimension of the spline, or -1.

/// @fn inline int EW_Dataflow_Spline3::lmk_index(int side, int point_set,
///  int i) const
/// To construct the spline, landmarks in the template and specimen are matched
/// by id and by position within the point_set.
/// This calculates the index in the spline of a point in a point_set.
/// @param side
///   0 if the point is in the template, 1 if it is in the specimen.
/// @param point_set
///   The index of the point_set.
/// @param i
///   The position of the point within the point_set.
/// @return
///   The index, if the point is matched, otherwise -1.

/// @fn inline int EW_Dataflow_Spline3::lmk_pointset(int side, int index) const
/// This is the inverse of #lmk_index.
/// @param side
///   0 for the pointset in the template, 1 for the pointset in the specimen.
/// @param index
///   The index as returned by #lmk_index.
/// @return
///   The pointset of the specified landmark in the spline.

/// @fn inline int EW_Dataflow_Spline3::lmk_pointset_i(
///  int side, int index) const
/// This is the inverse of #lmk_index.
/// @param side
///   0 for the pointset in the template, 1 for the pointset in the specimen.
/// @param index
///   The index as returned by #lmk_index.
/// @return
///   The index of the landmark in the pointset of the specified landmark in
///   the spline.

/// @fn inline bool EW_Dataflow_Spline3::get_nonsingular() const
/// @return
///   @c true if the spline is non-singular.

/// @fn inline double EW_Dataflow_Spline3::get_energy() const
/// If the spline is singular, a std::runtime_error is thrown.
/// @return
///   The bending energy of the spline.

/// @fn const double *EW_Dataflow_Spline3::get_optimized_lmk_images() const
/// This returns the positions of the landmarks after being allowed to slide
/// in their relaxation spaces to the positions that minimize bending energy.
/// The order of the landmarks is as defined by #lmk_index.
/// If the spline is singular, a std::runtime_error is thrown.
/// The pointer is valid until the node or a dependency of the node is
/// changed.
/// @return
///   A pointer to an array of coordinates.

/// @fn void EW_Dataflow_Spline3::warp_points(double *outp, const double *inp,
///  int l) const
/// This applies the spline to an array of points.
/// If the spline is singular, a std::runtime_error is thrown.
/// @param[out] outp
///   Where to store the warped point coordinates.
/// @param[in] inp
///   Where the original point coordinates are stored.
/// @param[in] l
///   The number of points.

// The rest of this file is private.

inline const EW_Dataflow_Form3 *
EW_Dataflow_Spline3::get_template() const
{
  return static_cast<const EW_Dataflow_Form3 *>(
   EW_Dataflow_Node::InpDependencies[0]);
}

inline const EW_Dataflow_Form3 *
EW_Dataflow_Spline3::get_specimen() const
{
  return static_cast<const EW_Dataflow_Form3 *>(
   EW_Dataflow_Node::InpDependencies[1]);
}

inline EW_Dataflow_Network::cycle_t
EW_Dataflow_Spline3::get_change_cycle_template() const
{
  return EW_Dataflow_Node::change_cycle_dependencies[0];
}

inline EW_Dataflow_Network::cycle_t
EW_Dataflow_Spline3::get_change_cycle_specimen() const
{
  return EW_Dataflow_Node::change_cycle_dependencies[1];
}

inline EW_Dataflow_Network::cycle_t
EW_Dataflow_Spline3::get_version_association() const
{
  EW_Dataflow_Node::update();
  return OutpVersionAssociation;
}

inline EW_Dataflow_Network::cycle_t
EW_Dataflow_Spline3::get_version_interaction() const
{
  EW_Dataflow_Node::update();
  return OutpVersionInteraction;
}

inline EW_Dataflow_Network::cycle_t
EW_Dataflow_Spline3::get_version_factorization() const
{
  EW_Dataflow_Node::update();
  return OutpVersionFactorization;
}

inline EW_Dataflow_Network::cycle_t
EW_Dataflow_Spline3::get_version_spline() const
{
  EW_Dataflow_Node::update();
  return OutpVersionSpline;
}

inline int
EW_Dataflow_Spline3::get_n_lmks() const
{
  make_association();
  return OutpNLmks;
}

inline int
EW_Dataflow_Spline3::get_f_size() const
{
  make_association();
  return OutpFSize;
}

inline int
EW_Dataflow_Spline3::lmk_index(int side, int point_set, int i) const
{
  make_association();
  int r = OutpPsToSpline[side][point_set];
  if (r < 0) {
    return -1;
  } else {
    return r + i;
  }
}

inline int
EW_Dataflow_Spline3::lmk_pointset(int side, int lmk_index) const
{
  make_association();
  if (lmk_index >= OutpNLmks) {
    return -1;
  }
  return OutpSplineToPs[side][lmk_index];
}

inline int
EW_Dataflow_Spline3::lmk_pointset_i(int side, int lmk_index) const
{
  make_association();
  if (lmk_index >= OutpNLmks) {
    return -1;
  }
  return OutpSplineToPsI[side][lmk_index];
}

inline bool
EW_Dataflow_Spline3::get_nonsingular() const
{
  make_factorization();
  return OutpNonsingular;
}

inline double
EW_Dataflow_Spline3::get_energy() const
{
  make_spline();
  return OutpEnergy;
}

inline const double *
EW_Dataflow_Spline3::get_optimized_lmk_images() const
{
  make_spline();
  return OutpRelaxLmkImages.empty() ? 0 : &OutpRelaxLmkImages[0] ;
}

inline void
EW_Dataflow_Spline3::make_association() const
{
  EW_Dataflow_Network::cycle_t v = get_version_association();
  if (update_cycle_association < v) {
    update_association();
    update_cycle_association = EW_Dataflow_Node::network->get_cycle();
  }
}

inline void
EW_Dataflow_Spline3::make_interaction() const
{
  EW_Dataflow_Network::cycle_t v = get_version_interaction();
  if (update_cycle_interaction < v) {
    make_association();
    update_interaction();
    update_cycle_interaction = EW_Dataflow_Node::network->get_cycle();
  }
}

inline void
EW_Dataflow_Spline3::make_factorization() const
{
  EW_Dataflow_Network::cycle_t v = get_version_factorization();
  if (update_cycle_factorization < v) {
// XXX probably unnecessary
    make_association();
// XXX probably unnecessary
    make_interaction();
    update_factorization();
    update_cycle_factorization = EW_Dataflow_Node::network->get_cycle();
  }
}

inline void
EW_Dataflow_Spline3::make_spline() const
{
  EW_Dataflow_Network::cycle_t v = get_version_spline();
  if (update_cycle_spline < v) {
// XXX probably unnecessary
    make_association();
// XXX probably unnecessary
    make_interaction();
// XXX probably unnecessary
    make_factorization();
    update_spline();
    update_cycle_spline = EW_Dataflow_Node::network->get_cycle();
  }
}

#endif
