// wdkg 2008-2010

#ifndef EW_DATAFLOW_SURFACE3_H_INCLUDED
#define EW_DATAFLOW_SURFACE3_H_INCLUDED 1

#include <memory>
#include "ew/bbox3.h"
#include "ew/dataflow_network.h"
#include "ew/dataflow_node.h"

class EWIndex3;

//XXX separate dlist for faces for use by warped surface
//XXX revert to generating dlist here as a update_ method (cf notes_nodes)
//XXX mechanism for optionally responding to file changes
//XXX   check for file modification when load, function for global reload
//XXX check dlist freed when no items using it

/// Surface Node Base Class

/// EW_Dataflow_Surface3 is the base class for nodes containing triangulated
/// surfaces in @f$\mathbb{R}^3@f$.
///
/// EW_Dataflow_Surface3 is a class without assignment or comparison.
/// There are private member variables.
///
/// EW_Dataflow_Surface3 is declared in @c ew/dataflow_surface3.h.
///
/// For efficient rendering of the same surface in multiple windows, an OpenGL
/// display list is constructed containing the basic surface geometry commands.
/// All windows displaying a given surface must share display lists.

class EW_Dataflow_Surface3 : public EW_Dataflow_Node {
public:
  inline int get_num_points() const;
  inline int get_num_faces() const;
  inline const float *get_points() const;
  inline const int *get_faces() const;
  inline const EW_Bbox3 *get_bbox() const;
  inline void make_index() const;
  inline bool index_is_made() const;
  double project(int *face, double *coeffs, double *normal, double *proj,
   double *inp) const;
protected:
  EW_Dataflow_Surface3(EW_Dataflow_Network *nw, int ndeps, const char *n);
  ~EW_Dataflow_Surface3();
  mutable int OutpNumPoints;
  mutable int OutpNumFaces;
  mutable const float *OutpPoints;
  mutable const int *OutpFaces;
  void reset_normals() const;
  void reset_index() const;
protected:
//XXX make these private (needed for reset)
  mutable EW_Dataflow_Network::cycle_t update_cycle_index;
  mutable EW_Dataflow_Network::cycle_t update_cycle_normals;
private:
  friend class EW_View3_Surface;
// EW_View3_Surface
  inline unsigned int get_dlist() const;
  inline void set_hold_dlist(unsigned int dl) const;
  inline void hold_dlist() const;
  inline void release_dlist() const;
  inline EW_Dataflow_Network::cycle_t get_dlist_cycle() const;
  inline void set_dlist_cycle(EW_Dataflow_Network::cycle_t cycle) const;
// private
  void update_bbox() const;
  inline const float *get_normals() const;
  void update_normals() const;
  void update_index() const;
  mutable EW_Bbox3 OutpBbox;
  mutable std::vector<float> OutpNormals;
  const std::auto_ptr<EWIndex3> OutpIndex;
  mutable EW_Dataflow_Network::cycle_t update_cycle_bbox;
  mutable unsigned int dlist;
  mutable EW_Dataflow_Network::cycle_t dlist_cycle;
  mutable int dlist_refs;
  const bool DbgOnSur;
};

// index is a pointer just to avoid having to define EWIndex3 in the
// exported header files.

/// @fn int EW_Dataflow_Surface3::get_num_points() const
/// This will force the network into an updating phase.
/// @return
///   The number of points used in the surface.

/// @fn int EW_Dataflow_Surface3::get_num_faces() const
/// This will force the network into an updating phase.
/// @return
///   The number of triangular faces in the surface.

/// @fn const float *EW_Dataflow_Surface3::get_points() const
/// This will force the network into an updating phase.
/// The pointer is valid until the node or a dependency of the node is changed.
/// @return
///   A pointer to the coordinates of the points used in the surface.

/// @fn const int *EW_Dataflow_Surface3::get_faces() const
/// This will force the network into an updating phase.
/// The pointer is valid until the node or a dependency of the node is changed.
/// @return
///   A pointer to the indices of the points used in the surface.

/// @fn const EW_Bbox3 *EW_Dataflow_Surface3::get_bbox() const
/// The pointer is valid until the node or a dependency of the node is changed.
/// @return
///   A pointer to the bounding box of the surface.

/// @fn inline void EW_Dataflow_Surface3::make_index() const
/// This makes the surface spatial index if it has not already been made since
/// the surface was last changed.
/// The spatial index is used by #project.

/// @fn inline bool EW_Dataflow_Surface3::index_is_made() const;
/// @return
///   @c true if the surface spatial index is up to date.

/// @fn double EW_Dataflow_Surface3::project(int *face, double *coeffs,
///  double *normal, double *proj, double *inp) const
/// This finds the nearest point on the surface to a given point.
/// The surface spatial index will be made if it has not already been made.
/// This might be time consuming.
/// The timing of this delay can be controlled by calling #make_index
/// beforehand.
/// @param[out] face
///   The index of the face of the surface containing the nearest point.
/// @param[out] coeffs
///   The coefficients of the nearest point when expressed as a linear
///    combination of the vertices of the face.
/// @param[out] normal
///   The interpolated normal at the nearest point.
///   If a sensible normal cannot be calculated, an arbitrary unit vector
///   is returned.
/// @param[out] proj
///   The coordinates of the nearest point on the surface.
/// @param[in] inp
///   The coordinates of the original point.
/// @return
///   The distance from the original point to the projected point.

// The rest of this file is private.

inline int
EW_Dataflow_Surface3::get_num_points() const
{
  EW_Dataflow_Node::update();
  return OutpNumPoints;
}

inline int
EW_Dataflow_Surface3::get_num_faces() const
{
  EW_Dataflow_Node::update();
  return OutpNumFaces;
}

inline const float *
EW_Dataflow_Surface3::get_points() const
{
  EW_Dataflow_Node::update();
  return OutpPoints;
}

inline const int *
EW_Dataflow_Surface3::get_faces() const
{
  EW_Dataflow_Node::update();
  return OutpFaces;
}

inline const EW_Bbox3 *
EW_Dataflow_Surface3::get_bbox() const
{
  EW_Dataflow_Network::cycle_t v = EW_Dataflow_Node::get_version();
  if (update_cycle_bbox < v) {
    update_bbox();
    update_cycle_bbox = EW_Dataflow_Node::network->get_cycle();
  }
  return &OutpBbox;
}

inline const float *
EW_Dataflow_Surface3::get_normals() const
{
  EW_Dataflow_Network::cycle_t v = EW_Dataflow_Node::get_version();
  if (update_cycle_normals < v) {
    update_normals();
    update_cycle_normals = EW_Dataflow_Node::network->get_cycle();
  }
  return OutpNormals.empty() ? 0 : &(OutpNormals[0]);
}

inline void
EW_Dataflow_Surface3::make_index() const
{
  EW_Dataflow_Network::cycle_t v = EW_Dataflow_Node::get_version();
  if (update_cycle_index < v) {
    update_index();
    update_cycle_index = EW_Dataflow_Node::network->get_cycle();
  }
}

inline bool
EW_Dataflow_Surface3::index_is_made() const
{
  return (update_cycle_index >= EW_Dataflow_Node::get_version());
}

inline unsigned int
EW_Dataflow_Surface3::get_dlist() const
{
  return dlist;
}

inline void
EW_Dataflow_Surface3::set_hold_dlist(unsigned int dl) const
{
  dlist = dl;
  dlist_cycle = 0;
  dlist_refs++;
}

inline EW_Dataflow_Network::cycle_t
EW_Dataflow_Surface3::get_dlist_cycle() const
{
  return dlist_cycle;
}

inline void
EW_Dataflow_Surface3::set_dlist_cycle(EW_Dataflow_Network::cycle_t cycle) const
{
  dlist_cycle = cycle;
}

inline void
EW_Dataflow_Surface3::hold_dlist() const
{
  dlist_refs++;
}


inline void
EW_Dataflow_Surface3::release_dlist() const
{
  dlist_refs--;
  if (dlist_refs == 0) {
    dlist = 0;
    dlist_cycle = 0;
  }
}

#endif
