// wdkg 2008-2010

#ifndef EW_DIG3_H_INCLUDED
#define EW_DIG3_H_INCLUDED 1

#include <string>
#include <vector>

class EW_Dataflow_Network;
class EW_Dataflow_Spline3;
class EW_Dig3_Tableau;
class EW_Dig3_Space;
class EW_Dig3_View;
class EWDebug;

// XXX for now, forms must be empty or have a filename (last load and save,
// not reset, but changes ok) before save records can be recorded

/// Digitizing Application

/// EW_Dig3 implements a 3D digitizing application.
///
/// EW_Dig3 is a class without assignment or comparison.
/// There are private member variables.
///
/// EW_Dig3 is declared in ew/dig3.h.
///
/// An EW_Dig3 instance maintains 2 space elements, the template and the
/// specimen.
///
/// View elements can be added and removed by creating or destroying
/// EW_Dig3_View instances.
/// The views are implemented as EW_View3_Widgets, but EW_Dig3 does not create
/// them.
/// They must be instances of a class derived from EW_View3_Widget that
/// implements the toolkit and gui interactions, and so must be created
/// by the application.

//XXX save/load state must be done in conjunction with save/load form
// latter not done automatically
// so dig3 does not have to deal with filenames
//XXX later maybe ? save file dirty or version
//XXX later maybe partial save file read/write
//XXX ? save -> state
//Full EW_Dig state consists of
//  specimen_form template_form
//  views, and their spaces, slice modes and links
//  save file (primary main and primary slice configuration)
//  configurations of other windows
//Subsiduary main/slice windows can have different on/off's but won't be
//recorded in save file.
//e.g may want to duplicate main window with surfaces transparent

class EW_Dig3 {
public:
  enum TableauFlags {
/// Set the view mapping.
    TABLEAU_VIEW = 0x01,
/// Set the settings.
    TABLEAU_SETTINGS = 0x02,
/// Set everything.
    TABLEAU_ALL = 0x03
  };
  explicit EW_Dig3(EW_Dataflow_Network *network);
  ~EW_Dig3();
  inline int get_n_views() const;
  inline EW_Dig3_View *const *get_views() const;
  inline EW_Dig3_Space *const *get_spaces() const;
  inline const EW_Dataflow_Spline3 *get_spline_node() const;
  void save_tableau(EW_Dig3_Tableau *outp,
   int template_main, int template_slice,
   int specimen_main, int specimen_slice);
//XXX ?could put this in EW_Dig3_View
  void load_tableau(const EW_Dig3_Tableau *inp, int view, unsigned int flags);
//XXX ??? do we allow file surface records in tableau but not in form file
//XXX     (tableau/form mismatch)
//XXX      would allow save file just to specify colour schemes, load form later
//XXX      instead of a speciific savefile/formfile combination
  void interpolate_tableau(const EW_Dig3_Tableau *inp1,
   const EW_Dig3_Tableau *inp2, double e, int view);
  EW_Dataflow_Network *const network;
private:
  friend class EW_Dig3_Space;
  friend class EW_Dig3_View;
// EW_Dig3_Space
  EW_Dig3_Space *spaces[2];
  EW_Dataflow_Spline3 *spl;
  int DbgLastSpaceN;
// EW_Dig3_View
  std::vector<EW_Dig3_View *> views;
  int DbgLastViewN;
// private
  const EWDebug &Dbg;
  const bool DbgOn;
//XXX review
  void dbg_check(const char *classname, int line);
public:
// private to library
  const int DbgN;
private:
  const std::string DbgStr;
public:
// private to library
  const char *const DbgId;
private:
// undefined
  EW_Dig3(const EW_Dig3 &rhs);
  EW_Dig3 &operator= (const EW_Dig3 &rhs);
};

/// @enum EW_Dig3::TableauFlags
/// These are flags that can be or-ed and passed to #load_tableau.

/// @fn explicit EW_Dig3::EW_Dig3(EW_Dataflow_Network *network)
/// This creates an empty digitizing application.
/// @param network
///   A pointer to a dataflow network instance.

/// @fn EW_Dig3::~EW_Dig3()
/// This destroys the digitizing application.
/// EW_Dig3_View widgets associated with it are destroyed.

/// @fn int EW_Dig3::get_n_views() const
/// @return
///   The number of views.

/// @fn EW_Dig3_View *const *EW_Dig3::get_views() const
/// @return
///   The view index.

/// @fn EW_Dig3_Space *const *EW_Dig3::get_spaces() const
/// @return
///   The space index.
///   There are always 2 spaces and the index is constant.

/// @fn inline const EW_Dataflow_Spline3 *EW_Dig3::get_spline_node() const
/// @return
///    A pointer to the spline node that this #EW_Dig3 manages.

/// @fn void EW_Dig3::save_tableau(EW_Dig3_Tableau *outp,
///  int template_main, int template_slice,
///  int specimen_main, int specimen_slice)
/// This creates a tableau record from the indicated views, specified by index.
/// The form filename fields of the tableau are left blank, and must be
/// explicitly set.
/// @param outp
///   Where to store the tableau record.
/// @param template_main
///   The index of a template main view to record.
/// @param template_slice
///   The index of a template slice view to record.
/// @param specimen_main
///   The index of a specimen main view to record.
/// @param specimen_slice
///   The index of a specimen slice view to record.

/// @fn void EW_Dig3::load_tableau(const EW_Dig3_Tableau *inp, int view,
///  unsigned int flags)
/// This applies a tableau record to the indicated view.
/// The type of view and its space determine which part of the record is
/// applied.
/// The form filename fields are not applied.
/// The forms must be explicitly set beforehand.
/// @param inp
///   The tableau to apply.
/// @param view
///   The index of the view to apply it to.
/// @param flags
///   Flags of type #TableauFlags indicating which parts of the tableau record
///   to apply.

/// @fn void EW_Dig3::interpolate_tableau(const EW_Dig3_Tableau *inp1,
///  const EW_Dig3_Tableau *inp2, double e, int view)
/// This interpolates between tableau records and applies the result to the
/// indicated view.
/// Only the views are applied.
/// The settings of the tableau and the forms cannot be interpolated and need
/// to be manually applied when a new frame is reached in a filmstrip.
/// @param inp1, inp2
///   The tableaus to interpolate.
/// @param e
///   The interpolation parameter, with 0.0 corresponding to @a inp1 and 1.0
///   to @a inp.
///   @a e is not restricted to [0.0, 1.0].
/// @param view
///   The index of the view to apply it to.

/// @var EW_Dataflow_Network *const EW_Dig3::network
/// This is a pointer to the EW_Dataflow_Network that this is associated with.

// The rest of this file is private.

inline EW_Dig3_Space *const *
EW_Dig3::get_spaces() const
{
  return spaces;
}

inline const EW_Dataflow_Spline3 *
EW_Dig3::get_spline_node() const
{
  return spl;
}

inline int
EW_Dig3::get_n_views() const
{
  return views.size();
}

inline EW_Dig3_View *const *
EW_Dig3::get_views() const
{
  return views.empty() ? 0 : &views[0];
}

#endif
