// wdkg 2008-2010

#ifndef EW_DIG3_TABLEAU_SPACE_H_INCLUDED
#define EW_DIG3_TABLEAU_SPACE_H_INCLUDED 1

#include <string>
#include "ew/dig3_tableau_curve.h"
#include "ew/dig3_tableau_surface.h"
#include "ew/transform3.h"

/// Space Element

/// EW_Dig3_Tableau_Space contains data for a space element of EW_Dig3_Tableau.
///
/// EW_Dig3_Tableau_Space is a data structure that supports default
/// construction, copy construction, assignment and equality comparison.
///
/// EW_Dig3_Tableau_Space is declared in @c ew/dig3_tableau_space.h.

class EW_Dig3_Tableau_Space {
public:
  std::string form_filename;
  EW_Transform3 main_view;
  EW_Transform3 slice_view;
  bool show_slice_in_main;
  bool show_lmks_in_main;
  bool show_lmks_in_slice;
  int lmks_symbol;
  unsigned char lmks_col[3];
  std::vector<EW_Dig3_Tableau_Curve> curve_settings;
  std::vector<EW_Dig3_Tableau_Surface> surface_settings;
  EW_Dig3_Tableau_Space();
  inline bool operator==(const EW_Dig3_Tableau_Space &a) const;
  inline bool operator!=(const EW_Dig3_Tableau_Space &a) const;
};

/// @var std::string EW_Dig3_Tableau_Space::form_filename
/// This file containing the form data for the space.

/// @var EW_Transform3 EW_Dig3_Tableau_Space::main_view
/// The view mapping of the main view of this space.
/// The default is the identity transformation.

/// @var EW_Transform3 EW_Dig3_Tableau_Space::slice_view
/// The view mapping of the slice view of this space.
/// The default is the identity transformation.

/// @var bool EW_Dig3_Tableau_Space::show_slice_in_main
/// @c true if the slice should be displayed in the main view.
/// The default is @c true.

/// @var bool EW_Dig3_Tableau_Space::show_lmks_in_main
/// @c true if the landmarks should be displayed in the main view.
/// The default is @c true.

/// @var bool EW_Dig3_Tableau_Space::show_lmks_in_slice
/// @c true if the landmarks should be displayed in the slice view.
/// The default is @c true.

/// @var int EW_Dig3_Tableau_Space::lmks_symbol
/// The symbol to use for landmarks.
/// The default is 0.

/// @var unsigned char EW_Dig3_Tableau_Space::lmks_col[3]
/// The colour of the landmarks.
/// The default is [0, 255, 0].

// Doxygen comment without long line or g++ multiline comment warning.
/*!
 * @var std::vector<EW_Dig3_Tableau_Curve> \
 *  EW_Dig3_Tableau_Space::curve_settings
 * The settings for curves in the space.
 */

// Doxygen comment without long line or g++ multiline comment warning.
/*!
 * @var std::vector<EW_Dig3_Tableau_Surface> \
 *  EW_Dig3_Tableau_Space::surface_settings
 * The settings for surfaces in the space.
 */

/// @fn EW_Dig3_Tableau_Space::EW_Dig3_Tableau_Space()
/// This sets intrinsic and POD member variables to their default values.

/// @fn inline bool EW_Dig3_Tableau_Space::operator==(
///  const EW_Dig3_Tableau_Space &a) const
/// This compares member by member.

/// @fn inline bool EW_Dig3_Tableau_Space::operator!=(
///  const EW_Dig3_Tableau_Space &a) const
/// This compares member by member.

// The rest of this file is private.

inline
EW_Dig3_Tableau_Space::EW_Dig3_Tableau_Space() :
 main_view(EW_Transform3::identity_transform),
 slice_view(EW_Transform3::identity_transform),
 show_slice_in_main(true),
 show_lmks_in_main(true),
 show_lmks_in_slice(true),
 lmks_symbol(0)
{
  lmks_col[0] = 0;
  lmks_col[1] = 255;
  lmks_col[2] = 0;
}

inline bool
EW_Dig3_Tableau_Space::operator==(const EW_Dig3_Tableau_Space &a) const
{
  return
   form_filename == a.form_filename &&
   main_view == a.main_view &&
   slice_view == a.slice_view &&
   show_slice_in_main == a.show_slice_in_main &&
   show_lmks_in_main == a.show_lmks_in_main &&
   show_lmks_in_slice == a.show_lmks_in_slice &&
   lmks_symbol == a.lmks_symbol &&
   lmks_col[0] == a.lmks_col[0] &&
   lmks_col[1] == a.lmks_col[1] &&
   lmks_col[2] == a.lmks_col[2] &&
   curve_settings == a.curve_settings &&
   surface_settings == a.surface_settings;
}

inline bool
EW_Dig3_Tableau_Space::operator!=(const EW_Dig3_Tableau_Space &a) const
{
  return !operator==(a);
}

#endif
