// wdkg 2008-2010

#ifndef EW_FORM3_H_INCLUDED
#define EW_FORM3_H_INCLUDED 1

#include <stdexcept>
#include <vector>
#include "ew/form3_curve.h"
#include "ew/form3_embedding.h"
#include "ew/form3_pointset.h"
#include "ew/form3_reflection.h"
#include "ew/form3_surface.h"
#include "ew/form3_volume.h"

//XXX doc well-formedness and consequences of ill-formedness
//XXX cannot have more than 2G of anything or of bytes: doc, check and throw
//XXX ? don't order, use linear search (spline association slow)
//XXX ? maintain order of pointsets in EW_Dataflow_Form3
//XXX check and canonicalize method:
//XXX   ordered id's
//XXX   id's non-empty and distinct
//XXX   embedding and reflection not hanging and appropriate dims/reflexivity
//XXX   export order a permutation
//XXX check and canonicalize on read
//XXX transform/stretch in volume
//XXX export order (landmarks only?) (maintain on insert/delete)

/// Morphometric Form

/// EW_Form3 represents morphometric forms in @f$\mathbb{R}^3@f$.
/// These are landmark configurations, possibly including semi-landmarks,
/// generalized to include curve, surface and volume data.
///
/// EW_Form3 is a data structure that supports default construction, copy
/// construction, assignment and equality comparison.
///
/// EW_Form3 is declared in @c ew/form3.h.
///
/// EW_Form3 data is represented in form files read by #read_file and written
/// by #write_file.
/// Here are some example files:
///   - @ref simple_landmarks.form
///   - @ref mandible_template.form
///   - @ref mandible_case.form
/// The representation is exact except that filenames in the form files are
/// abbreviated to relative filenames in common situations were possible, but
/// in EW_Form3 they are always in canonical form.
///
/// All elements in the form have an id.
/// These id's must be unique in the from.

class EW_Form3 {
public:
  enum StateType {
    STATE_SET,
    STATE_UNSET,
    STATE_PROVISIONAL,
    STATE_WARPED,
    STATE_PROJECTED,
    STATE_OPTIMIZED
  };
  static const int N_STATES = 6;
  static const char *const state_names[];
  enum PointType {
    TYPE_LANDMARK,
    TYPE_SEMI_LANDMARK,
    TYPE_POINT,
    TYPE_LINE,
    TYPE_PLANE,
    TYPE_FRAME
  };
  static const int N_TYPES = 6;
  static const char *const type_names[];
  std::vector<EW_Form3_Volume> volumes;
  std::vector<EW_Form3_Surface> surfaces;
  std::vector<EW_Form3_Curve> curves;
  std::vector<EW_Form3_PointSet> pointsets;
  std::vector<EW_Form3_Embedding> embeddings;
  std::vector<EW_Form3_Reflection> reflections;
  void reset();
  void read_file(const char *file);
  void write_file(const char *file, bool compress) const;
  bool search_volume(int *position, const char *id) const;
  bool search_surface(int *position, const char *id) const;
  bool search_curve(int *position, const char *id) const;
  bool search_pointset(int *position, const char *id) const;
  bool search_embedding(int *position, const char *id1, const char *id2) const;
  bool search_reflection(int *position, const char *id1,
   const char *id2) const;
  inline void swap(EW_Form3 &f);
  bool operator==(const EW_Form3 &a) const;
  inline bool operator!=(const EW_Form3 &a) const;
};

inline void swap(EW_Form3 &f1, EW_Form3 &f2);

/// @enum EW_Form3::StateType
/// Some elements can have a state associated with them.
/// These are the possible values.
/// Not all states are appropriate for all element types.
/// The default is #STATE_SET.

/// @var EW_Form3::STATE_SET
/// The element has been explicitly digitized (default).

/// @var EW_Form3::STATE_UNSET
/// The element has not been digitized.

/// @var EW_Form3::STATE_PROVISIONAL
/// The element has been provisionally digitized.

/// @var EW_Form3::STATE_WARPED
/// The element is the image under a warp.

/// @var EW_Form3::STATE_PROJECTED
/// The element has been projected.

/// @var EW_Form3::STATE_OPTIMIZED
/// The element has been optimized.

/// @enum EW_Form3::PointType
/// Point sets can have different types.
/// These are the possible values.
/// The default is #TYPE_LANDMARK.

/// @var EW_Form3::TYPE_LANDMARK
/// The points are landmarks (default).

/// @var EW_Form3::TYPE_SEMI_LANDMARK
/// The points are semi-landmarks.

/// @var EW_Form3::TYPE_POINT
/// The point is a non-landmark point.

/// @var EW_Form3::TYPE_LINE
/// The point represents a line, such as a best-fit axis of symmetry.

/// @var EW_Form3::TYPE_PLANE
/// The point represents a plane, such as a best-fit multi-tangent
/// or best-fit symmetry plane.

/// @var EW_Form3::TYPE_FRAME
/// The point represents a frame, such as a standard view.

/// @var std::vector<EW_Form3_Volume> EW_Form3::volumes
/// An arbitrary number of volume elements.
/// The volume data is stored in separate files.
/// The files named are not expected to change during their use.
/// These must be in alphabetical order by EW_Form3_Volume::id in the POSIX
/// sorting order.

/// @var std::vector<EW_Form3_Surface> EW_Form3::surfaces
/// An arbitrary number of surface elements.
/// The surface data is stored in separate files.
/// The files named are not expected to change during their use.
/// These must be in alphabetical order by EW_Form3_Surface::id in the POSIX
/// sorting order.

/// @var std::vector<EW_Form3_Curve> EW_Form3::curves
/// An arbitrary number of curve elements.
/// The curve data is stored in separate files.
/// The files named can be created during the digitization, so can change
/// during their use.
/// These must be in alphabetical order by EW_Form3_Curve::id in the POSIX
/// sorting order.

/// @var std::vector<EW_Form3_PointSet> EW_Form3::pointsets
/// An arbitrary number of point set elements.
/// These must be in alphabetical order by id in the POSIX sorting order.

/// @var std::vector<EW_Form3_Embedding> EW_Form3::embeddings
/// An arbitrary number of individual embedding relationships.
/// These must be in alphabetical order by EW_Form3_Embedding::subset_id,
/// then EW_Form3_Embedding::superset_id in the POSIX sorting order.
/// They must refer to elements of the form of appropriate dimension.
/// This is how the curve or surface a semi-landmark belongs to is indicated.
/// The semi-landmarks must be embedded in a unique curve or surface of the
/// appropriate dimension.
/// Regular landmarks can be embedded in one or more curves or surfaces.

/// @var std::vector<EW_Form3_Reflection> EW_Form3::reflections
/// An arbitrary number of individual reflection relationships.
/// These must be in alphabetical order by EW_Form3_Reflection::left_id,
/// then EW_Form3_Reflection::right_id in the POSIX sorting order.
/// They must refer to elements of the form of appropriate dimension.

/// @fn void EW_Form3::reset()
/// This resets the form to its initial state.

/// @fn void EW_Form3::read_file(const char *file)
/// This reads a form from a file.
/// The file must be a real filename, URL's are not supported.
/// The file can be compressed with gzip compression.
/// Relative filenames with initial path component "." inside the file are
/// converted to canonical filenames using the directory part of @a file.
/// If an exception is thrown, the form is unchanged.
/// @param file
///   The canonical name of the file to read.
/// @exception std::bad_alloc
/// @exception EW_Form3_IOError

/// @fn void EW_Form3::write_file(const char *file, bool compress) const
/// This writes a form to a file.
/// Filenames written into the output are converted to relative filenames with
/// initial path component "." if their path contains the directory part of @a
/// file as an initial segment.
/// @param file
///   The canonical name of the file to write.
/// @param compress
///   If @c true, write the file compressed with gzip compression.
/// @exception std::bad_alloc
/// @exception EW_Form3_IOError

/// @fn bool EW_Form3::search_volume(int *position, const char *id) const
/// This searches the volumes in the form for an id.
/// @param[out] position
///   The index in the vector that a volume with this id would be if
///   inserted into the vector of volumes.
/// @param id
///   The id to search for.
/// @return
///   @c true if the id is an existing volume id.

/// @fn bool EW_Form3::search_surface(int *position, const char *id) const
/// This searches the surfaces in the form for an id.
/// @param[out] position
///   The index in the vector that a surface with this id would be if
///   inserted into the vector of surfaces.
/// @param id
///   The id to search for.
/// @return
///   @c true if the id is an existing surface id.

/// @fn bool EW_Form3::search_curve(int *position, const char *id) const
/// This searches the curves in the form for an id.
/// @param[out] position
///   The index in the vector that a curve with this id would be if
///   inserted into the vector of curves.
/// @param id
///   The id to search for.
/// @return
///   @c true if the id is an existing curve id.

/// @fn bool EW_Form3::search_pointset(int *position, const char *id) const
/// This searches the point sets in the form for an id.
/// @param[out] position
///   The index in the vector that a point set with this id would be if
///   inserted into the vector of point sets.
/// @param id
///   The id to search for.
/// @return
///   @c true if the id is an existing point set id.

/// @fn bool EW_Form3::search_embedding(int *position, const char *id1,
///  const char *id2) const
/// This searches the embeddings in the form for the id's.
/// @param[out] position
///   The index in the vector that a embedding with these id's would be if
///   inserted into the vector of embeddings.
/// @param id1
///   The subset_id to search for.
/// @param id2
///   The superset_id to search for.
/// @return
///   @c true if the id's belong to an existing embedding.

/// @fn bool EW_Form3::search_reflection(int *position, const char *id1,
///  const char *id2) const
/// This searches the reflections in the form for the id's.
/// @param[out] position
///   The index in the vector that a reflection with these id's would be if
///   inserted into the vector of reflections.
/// @param id1
///   The left_id to search for.
/// @param id2
///   The right_id to search for.
/// @return
///   @c true if the id's belong to an existing reflection.

/// @fn void EW_Form3::swap(EW_Form3 &f)
/// This swaps the data of 2 forms without any copying.
/// @param f
///   The form to swap with @c this.

/// @fn inline bool EW_Form3::operator==(const EW_Form3 &a) const
/// This compares member by member.

/// @fn inline bool EW_Form3::operator!=(const EW_Form3 &a) const
/// This compares member by member.

// The rest of this file is private.

inline void
EW_Form3::swap(EW_Form3 &f)
{
  volumes.swap(f.volumes);
  surfaces.swap(f.surfaces);
  curves.swap(f.curves);
  pointsets.swap(f.pointsets);
  embeddings.swap(f.embeddings);
  reflections.swap(f.reflections);
}

inline bool
EW_Form3::operator!=(const EW_Form3 &a) const
{
  return !operator==(a);
}

inline void
swap(EW_Form3 &f1, EW_Form3 &f2)
{
  f1.swap(f2);
}

#endif
