// wdkg 2008-2010

#ifndef EW_TPS2_H_INCLUDED
#define EW_TPS2_H_INCLUDED 1

/// Thin-Plate Spline in 2D

/// EW_Tps2 contains static methods that implement thin-plate spline
/// calculations in @f$\mathbb{R}^2@f$.
/// It supports semi-landmarks that are allowed to slide along affine
/// subspaces.
///
/// The method interfaces are purely procedural.
/// It is up to the caller to provide the space for the arguments and for the
/// results.
/// The only memory management performed is for internal temporary scratch
/// space.
///
/// EW_Tps2 is declared in @c ew/tps2.h.
///
/// The numerical linear algebra is performed by the LAPACK library.
/// The thin-plate spline system of equations has a symmetric, semi-definite
/// matrix of coefficients.
/// It is factorizated using the LAPACK function @c DSYTRF
/// and solve using the LAPACK function @c DSYTRS.
/// @c DSYTRF uses the Bunch-Kaufman diagonal pivoting algorithm.
/// In the calculation of bending energy, the matrix inversion is performed by
/// @c DSYTRI.
/// In the calculation of principal warps and principal axes, the eigenanalysis
/// is preformed by @c DSYEV.
/// Lapack includes an error handler, @c xerbla, which will get linked into the
/// program.
/// However, provided this class is used correctly, @c xerbla will never be
/// called.
///
/// On Mac OS X (as of 2010-02), the bundled lapack implementation can crash if
/// the array arguments are not 16 byte aligned.
/// Consequently, the array arguments of Tps2 methods should be 16 byte
/// aligned.
///
/// In this documentation, the first index of a matrix element is considered to
/// be the row number, the second the column number.
/// Matrix elements of all these methods are stored consecutively in memory by
/// rows.
///
/// See @ref EWsemilandmarks and @ref EWuniformwarps for details of the
/// algorithms used by these methods.

class EW_Tps2 {
public:
  static void interaction(
    double *L,
    const double *lmks,
    int n_lmks
  );
  static void algebraic_size(
    int *f_size,
    bool *is_mixed,
    bool *is_reduced,
    const int *relax_dims_opt,
    int n_lmks
  );
  static void factorize(
    double *F,
    int *pivots,
    double *nonsingularity,
    const double *L,
    const int *relax_dims_opt,
    const double *relax_params_opt,
    int f_size,
    bool is_mixed,
    bool is_reduced,
    int n_lmks
  );
  static void solve(
    double *spline,
    double *energy_opt,
    double *relax_lmk_images_opt,
    const double *F,
    const int *pivots,
    const double *lmk_images,
    const int *relax_dims_opt,
    const double *relax_params_opt,
    const double *L,
    int n_lmks,
    int f_size,
    bool is_mixed,
    bool is_reduced
  );
  static void map(
    double *point_images,
    const double *points,
    int n_points,
    const double *spline,
    const double *lmks,
    int n_lmks
  );
  static void cache(
    float *cache,
    const float *points,
    int n_points,
    const double *lmk
  );
  static void cache_map(
    float *point_images,
    const float *points,
    float *const (*caches),
    int n_points,
    const double *spline,
    int n_lmks
  );
  static void bending_energy(
    double *B,
    const double *F,
    const int *pivots,
    const int *relax_dims_opt,
    const double *relax_params_opt,
    int n_lmks,
    int f_size,
    bool is_mixed,
    bool is_reduced
  );
  static void principal_warps(
    double *P,
    double *evals,
    const double *B,
    int n_lmks,
    bool is_mixed
  );
  static void principal_axes(
    double *centroid,
    double *sum_squares,
    double *axes,
    const double *lmks,
    int n_lmks
  );
  static void uniform_basis(
    double *basis,
    double *transforms_opt,
    const double *centroid,
    const double *sum_squares,
    const double *axes,
    const double *lmks,
    int n_lmks
  );
private:
// undefined
  EW_Tps2();
  EW_Tps2(const EW_Tps2 &rhs);
  EW_Tps2 &operator=(const EW_Tps2 &rhs);
};

/// @fn void EW_Tps2::interaction(
///   double *L,
///   const double *lmks,
///   int n_lmks
/// )
/// This calculates the thin-plate spline interaction matrix of a configuration
/// of landmarks.
/// @param[out] L
///   This is a matrix of size (@a n_lmks + 3) x (@a n_lmks + 3).
///   The result is a symmetric and only the upper diagonal part is stored.
/// @param[in] lmks
///   This is a matrix of size @a n_lmks x 2 containing the landmark
///   coordinates.
/// @param[in] n_lmks
///   This is the number of landmarks.

/// @fn void EW_Tps2::algebraic_size(
///   int *f_size,
///   bool *is_mixed,
///   bool *is_reduced,
///   const int *relax_dims_opt,
///   int n_lmks
/// )
/// This calculates the algebraic dimension and other characteristics of a
/// thin-plate spline configuration with semi-landmarks.
/// It is not necessary to call this function if there are no semi-landmarks.
/// @param[out] f_size
///   This is the location to store the algebraic dimension of the thin-plate
///   spline equation.
///   @a f_size is needed to allocate space for arguments to other functions.
///   If @a is_mixed is @c false, it is at most @a n_lmks + 3.
///   Otherwise it is less than 2 * (@a n_lmks + 3).
///   If there are no semi-landmarks, it is exactly @a n_lmks + 3.
///   In general, the difference between these upper limits and @a f_size is
///   the sum of the elements of @a relax_dims.
/// @param[out] is_mixed
///   This is the location to store a @c boolean value indicating whether the 2
///   dimensions are mixed in the thin-plate spline equation.
///   If not, the thin-plate spline equation can be solved in each dimension
///   separately.
///   If there are no semi-landmarks, this will be @c false.
///   It is @c true if there are semi-landmarks sliding in spaces of non-zero
///   dimension and non-zero co-dimension.
/// @param[out] is_reduced
///   This is the location to store a @c boolean value indicating whether there
///   are any non-trivial semi-landmarks, and consequently, whether the
///   thin-plate spline equation has been reduced.
///   If there are no semi-landmarks, this will be @c false.
///   It is @c true if there are semi-landmarks sliding in spaces of non-zero
///   dimension.
/// @param[in] relax_dims_opt
///   This is an array of size @a n_lmks containing the dimensions of the
///   sliding spaces for each landmark.
///   Each can be 0, 1, or 2.
///   For landmarks that are not semi-landmarks, the entries should be 0.
///   If there are no semi-landmarks, this can be @c NULL.
/// @param[in] n_lmks
///   See #interaction.

/// @fn void EW_Tps2::factorize(
///   double *F,
///   int *pivots,
///   double *nonsingularity,
///   const double *L,
///   const int *relax_dims_opt,
///   const double *relax_params_opt,
///   int f_size,
///   bool is_mixed,
///   bool is_reduced,
///   int n_lmks
/// )
/// This calculates the algebraic equation corresponding to the semi-landmark
/// constraints, and factorizes the coefficient matrix.
/// If @a is_mixed is @c true, this equation is mixed.
/// Otherwise it separates into equations in each coordinate direction.
/// It uses outputs from #interaction and #algebraic_size.
/// @param[out] F
///   This is a matrix of size @a f_size x @a f_size.
///   The factorization of the coefficient matrix is stored here.
///   It is not output if the thin-plate spline is singular.
/// @param[out] pivots
///   This is an array of size @a f_size.
///   The pivots used in the factorization are stored here.
///   It is not output if the thin-plate spline is singular.
/// @param[out] nonsingularity
///   This is the inverse of the condition number of the block diagonal matrix
///   resulting from the matrix factorization.
///   If this is exactly 0.0, the other outputs will not have been set and they
///   must not used in #solve.
///   If this is close to 0.0, the other outputs are unreliable and should also
///   not be used.
///   A spline is exactly singular when there are 2 landmarks with the same
///   coordinates, or when the landmarks are colinear.
///   In this case, the @a nonsingularity might not be exactly 0.0 due to
///   numerical error.
///   Conversely, the spline equation might be numerically unsolvable without
///   it being exactly singular, for example, if there are an excessive number
///   of landmarks, if some landmarks are very close together, or if the
///   landmarks are nearly colinear.
///   The lower limit for the @a nonsingularity should be high enough to catch
///   these cases, but not so high as to reject too many usable splines.
///   A limit of @c 1.0e-8 is a sensible starting point.
/// @param[in] L
///   See #interaction.
/// @param[in] relax_dims_opt
///   See #algebraic_size.
/// @param[in] relax_params_opt
///   This is a matrix of size @a n_lmks x 2.
///   For every semi-landmark sliding along a line, the corresponding row of
///   this matrix should be a unit vector normal to the line.
///   The other rows need not be set.
///   If there are no semi-landmarks of dimension 1, this can be @c NULL.
/// @param[in] f_size
///   See #algebraic_size.
/// @param[in] is_mixed
///   See #algebraic_size.
/// @param[in] is_reduced
///   See #algebraic_size.
/// @param[in] n_lmks
///   See #interaction.
/// @exception std::bad_alloc

/// @fn void EW_Tps2::solve(
///   double *spline,
///   double *energy_opt,
///   double *relax_lmk_images_opt,
///   const double *F,
///   const int *pivots,
///   const double *lmk_images,
///   const int *relax_dims_opt,
///   const double *relax_params_opt,
///   const double *L,
///   int n_lmks,
///   int f_size,
///   bool is_mixed,
///   bool is_reduced
/// )
/// This calculates a thin-plate spline from a configuration of landmark
/// images.
/// It uses outputs from #interaction, #algebraic_size and #factorize.
/// It should not be called if the thin-plate spline is singular.
/// @param[out] spline
///   This is a matrix of size (@a n_lmks + 3) x 2.
///   The coefficients of the solution thin-plate spline are stored here.
/// @param[out] energy_opt
///   This is the location to store the bending energy of the thin-plate
///   spline.
///   It can be @c NULL, in which case the energy is not calculated.
/// @param[out] relax_lmk_images_opt
///   This is a matrix of size @a n_lmks x 2.
///   It is the location to store the optimized landmark images.
///   These are the landmark images that would minimize bending energy
///   consistent
///   with the semi-landmark constraints.
///   It can be @c NULL, in which case they are not calculated.
/// @param[in] F
///   See #factorize.
/// @param[in] pivots
///   See #factorize.
/// @param[in] lmk_images
///   This is a matrix of size @a n_lmks x 2 containing the landmark image
///   coordinates.
/// @param[in] relax_dims_opt
///   See #algebraic_size.
/// @param[in] relax_params_opt
///   See #factorize.
/// @param[in] L
///   See #interaction.
/// @param[in] n_lmks
///   See #interaction.
/// @param[in] f_size
///   See #algebraic_size.
/// @param[in] is_mixed
///   See #algebraic_size.
/// @param[in] is_reduced
///   See #algebraic_size.
/// @exception std::bad_alloc

/// @fn void EW_Tps2::map(
///   double *point_images,
///   const double *points,
///   int n_points,
///   const double *spline,
///   const double *lmks,
///   int n_lmks
/// )
/// This calculates the images of a list of points under a thin-plate spline.
/// It uses outputs of #solve.
/// It should not be called if the thin-plate spline is singular.
/// @param[in] point_images
///   This is a matrix of size @a n_points x 2.
///   It is the location to store the coordinates of the images of the points.
/// @param[in] points
///   This is a matrix of size @a n_points x 2, containing the coordinates of
///   the points
/// @param[in] n_points
///   This is the number of points.
/// @param[in] spline
///   See #solve.
/// @param[in] lmks
///   See #interaction.
/// @param[in] n_lmks
///   See #interaction.

/// @fn void EW_Tps2::cache(
///   float *cache,
///   const float *points,
///   int n_points,
///   const double *lmk
/// )
/// This evaluates a single thin-plate spline kernel function at a list of
/// points.
/// It is intended for speed critical situations where incremental changes
/// are made to a thin-plate spline that is being applied to a fixed set of
/// points.
/// It requires a @a cache for each landmark, which could be a require a
/// significant amount of memory.
/// The @a points and @a cache are of type @c float rather than @c double to
/// save space.
/// As a consequence, the accuracy of the mapped points is reduced.
/// @param[out] cache
///   This is an array of size n_points.
///   The cached evaluations are stored here.
/// @param[in] points
///   See #map, except that this is a @c float array rather than @c double.
/// @param[in] n_points
///   See #map.
/// @param[in] lmk
///   This points to an array of size 2 containing the coordinates of the
///   landmark.

/// @fn void EW_Tps2::cache_map(
///   float *point_images,
///   const float *points,
///   float *const (*caches),
///   int n_points,
///   const double *spline,
///   int n_lmks
/// )
/// This uses previously calculated cached kernel evaluations to calculate the
/// images of a list of points under a thin-plate spline.
/// It uses outputs from #solve and #cache.
/// It should not be called if the thin-plate spline is singular.
/// @param[in] point_images
///   See #map.
/// @param[in] points
///   See #map.
/// @param[in] caches
///   This is an array of size @a n_lmks.
///   It is an array of pointers, each pointing to the corresponding cache.
/// @param[in] n_points
///   See #map.
/// @param[in] spline
///   See #solve.
/// @param[in] n_lmks
///   See #interaction.

/// @fn void EW_Tps2::bending_energy(
///   double *B,
///   const double *F,
///   const int *pivots,
///   const int *relax_dims_opt,
///   const double *relax_params_opt,
///   int n_lmks,
///   int f_size,
///   bool is_mixed,
///   bool is_reduced
/// )
/// This calculates the bending energy matrix.
/// If @a is_mixed, it is a quadratic form on the mixed coordinates.
/// Otherwise, it is a quadratic form that is applied separately to the x
/// coordinates and to the y coordinates.
/// If there are any non-trivial semi-landmarks, this is the form for the
/// minimized bending energy.
/// It uses outputs from #interaction, #algebraic_size and #factorize.
/// It should not be called if the thin-plate spline is singular.
/// @param[out] B
///   This is the bending energy matrix.
///   It is a symmetric matrix, with both upper and lower diagonal entries
///   stored.
///   If @a is_mixed, it has size (2 * @a n_lmks) x (2 * @a n_lmks).
///   Otherwise, it has size @a n_lmks x @a n_lmks.
/// @param[in] F
///   See #factorize.
/// @param[in] pivots
///   See #factorize.
/// @param[in] relax_dims_opt
///   See #algebraic_size.
/// @param[in] relax_params_opt
///   See #factorize.
/// @param[in] n_lmks
///   See #interaction.
/// @param[in] f_size
///   See #algebraic_size.
/// @param[in] is_mixed
///   See #algebraic_size.
/// @param[in] is_reduced
///   See #algebraic_size.
/// @exception std::bad_alloc

/// @fn void EW_Tps2::principal_warps(
///   double *P,
///   double *evals,
///   const double *B,
///   int n_lmks,
///   bool is_mixed
/// )
/// This calculates the principal warps.
/// If @a is_mixed, they are a basis of the full space of landmark
/// displacements.
/// Otherwise, they are a basis of the space of landmarks in a single
/// coordinate directions, either x or y.
/// It uses outputs from #bending_energy.
/// It should not be called if the thin-plate spline is singular.
/// @param[out] P
///   If @a is_mixed is @c false, this is a matrix of size
///   @a n_lmks x @a n_lmks.
///   Otherwise it is a matrix of size (2 * @a n_lmks) x (2 * @a n_lmks).
///   The eigenvectors are stored here.
///   The form an orthogonal matrix.
/// @param[out] evals
///   If @a is_mixed is @c false, this is an array of size @a n_lmks.
///   Otherwise it is an array of size 2 * @a n_lmks.
///   The eigenvalues are stored here.
///   The eigenvectors are ordered so that the eigenvalues are in
///   non-decreasing order.
/// @param[in] B
///   See #bending_energy.
/// @param[in] n_lmks
///   See #interaction.
/// @param[in] is_mixed
///   See #algebraic_size.
/// @exception std::runtime_error
///   The eigenanalysis failed to converge.
/// @exception std::bad_alloc

/// @fn void EW_Tps2::principal_axes(
///   double *centroid,
///   double *sum_squares,
///   double *axes,
///   const double *lmks,
///   int n_lmks
/// )
/// This calculates a set of principal axes for the landmark configuration.
/// The axes are not canonical, which is why their calculation and their use
/// in #uniform_basis are separated into different functions.
/// They are canonical up to sign if the sum_squares are different.
/// It should not be called with an empty configuration of landmarks.
/// @param[out] centroid
///   This points to an array of size 2 to store the centroid of the
///   configuration.
/// @param[out] sum_squares
///   This points to an array of size 2 to store the sum of squared
///   coordinates in the principal axes.
/// @param[out] axes
///   This is a matrix of size @a n_lmks x @a n_lmks.
///   It is the location to store the principal axes.
///   They form an orthogonal matrix.
/// @param[in] lmks
///   See #interaction.
/// @param[in] n_lmks
///   See #interaction.
/// @exception std::runtime_error
///   The eigenanalysis failed to converge.
/// @exception std::bad_alloc

/// @fn void EW_Tps2::uniform_basis(
///   double *basis,
///   double *transforms_opt,
///   const double *centroid,
///   const double *sum_squares,
///   const double *axes,
///   const double *lmks,
///   int n_lmks
/// )
/// This calculates a canonical basis of the uniform subspace of the space of
/// shape variations, given a set of principal axes for the configuration.
/// The result is relative to the original axes, not the principal axes.
/// If the configuration is degenerate (contained within a line), this should
/// not be called.
/// @param[out] basis
///   This is a matrix of size 2 x @a n_lmks x 2.
///   The shape variables of the 2 basis elements are stored here.
///   The first index is the number of the basis element.
/// @param[out] transforms_opt
///   This points to an array of size 2 * 2 * 2.
///   Uniform shape variations are linear maps, so can be represented as 2 x 2
///   matrices.
///   The matrices of the 2 basis elements are stored here.
///   The first index is the number of the basis element.
///   This can be @c NULL, in which case the transformations are not
///   calculated.
/// @param[in] centroid
///   See #principal_axes.
/// @param[in] sum_squares
///   See #principal_axes.
/// @param[in] axes
///   See #principal_axes.
/// @param[in] lmks
///   See #interaction.
/// @param[in] n_lmks
///   See #interaction.

#endif
