// wdkg 2008-2010

#ifndef EW_TRANSFORM2_H_INCLUDED
#define EW_TRANSFORM2_H_INCLUDED 1

/// 2D Similarity Transformation

/// EW_Transform2 represents similarity transformations of @f$\mathbb{R}^2@f$,
/// each as the composition of an orthogonal linear map, a scaling and a
/// translation.
/// The translation is applied after the orthogonal linear map and the scaling.
///
/// EW_Transform2 is a POD type class.
///
/// EW_Transform2 is declared in @c ew/transform2.h.

class EW_Transform2 {
public:
  double orthog[2][2];
  double scale;
  double translate[2];
  unsigned int comps_cnt;
  void set_to_identity();
  void set_to_inverse(
    const EW_Transform2 *a
  );
  void set_to_composition(
    const EW_Transform2 *a,
    const EW_Transform2 *b
  );
  bool set_to_interpolation(
    const EW_Transform2 *a,
    const EW_Transform2 *b,
    double e
  );
  void get_matrix_gl(
    double *buffer
  ) const;
  inline void apply(
    float *dst,
    const float *src
  ) const;
  inline void apply(
    double *dst,
    const double *src
  ) const;
  double get_denormalization() const;
  void set_to_normalization(
    const EW_Transform2 *a
  );
  bool get_valid() const;
  bool operator==(const EW_Transform2 &a) const;
  inline bool operator!=(const EW_Transform2 &a) const;
  static const EW_Transform2 identity_transform;
};

/// @var double *EW_Transform2::orthog[2][2]
/// This is a matrix representing the orthogonal part of the transformation.
/// The matrix elements are ordered by row, and the matrix is applied on the
/// left.
/// The coordinates should be finite.
/// Deviations from orthogonaliity are tolerated.

/// @var double EW_Transform2::scale
/// This is the scale part of the transformation.
/// This should be positive.

/// @var double EW_Transform2::translate[2]
/// This is the translation part of the transformation.
/// The coordinates should be finite.

/// @var unsigned int EW_Transform2::comps_cnt
/// This is a count of the accumulated number of compositions performed in the
/// calculation of the current transformation.
/// It is reset by #set_to_identity and #set_to_normalization and set by
/// #set_to_inverse, #set_to_composition and #set_to_interpolation.
/// If the transformation is explicitly initialized, #comps_cnt should be
/// initialized to zero.
///
/// #set_to_composition will automatically normalize its result when #comps_cnt
/// exceeds an internal threshold.
///
/// This is to prevent exponential deviation of the orthogonal part from
/// orthogonality under iterated compositions.

/// @fn void EW_Transform2::set_to_identity()
/// This sets the transformation to the identity transformation.

/// @fn void EW_Transform2::set_to_inverse(
///   const EW_Transform2 *a
/// )
/// This sets the transformation to the inverse of another transformation.
/// @param a
///   This points to the transformation to invert.
///   @a a can equal @c this.

/// @fn void EW_Transform2::set_to_composition(
///   const EW_Transform2 *a,
///   const EW_Transform2 *b
/// )
/// This sets the transformation to the composition of two transformations.
/// @param a
///   This points to the transformation that is applied last in the
///   composition.
///   @a a can equal @c this.
/// @param b
///   This points to the transformation that is applied first in the
///   composition.
///   @a b can equal @c this.

/// @fn bool EW_Transform2::set_to_interpolation(
///   const EW_Transform2 *a,
///   const EW_Transform2 *b,
///   double e
/// )
/// This sets the transformation to an interpolation between two
/// transformations.
/// @param a
///   This points to the first transformation.
///   @a a can equal @c this.
/// @param b
///   This points to the second transformation.
///   @a b can equal @c this.
/// @param e
///   This is the interpolation paramater.
///   When 0.0, the transformation is set to @c *a, when 1.0 to @c *b.
///   @a e is not restricted to the range [0.0, 1.0].
/// @return
///   @c true if a smooth canonical interpolation was possible.
///   If @c false, the transformation is set to either the first or second
///   transformation, depending on whether @a e is closer to 0.0 or 1.0.
///   This can occur if one transformation is a reflection and the other
///   transformation is a rotation, or if the orthogonal parts differ by a half
///   turn.

/// @fn void EW_Transform2::get_matrix_gl(
///   double *buffer
/// ) const
/// This outputs the transformation as a 4 x 4 matrix in the format used by
/// @b OpenGL.
/// @param[out] buffer
///   The result is stored in this array of size 16.

/// @fn void EW_Transform2::apply(
///   float *dst,
///   const float *src
/// ) const
/// This applies the transformation to a point.
/// @param[out] dst
///   The result is stored in this array of size 2.
/// @param[in] src
///   The original point is contained in this array of size 2.

/// @fn void EW_Transform2::apply(
///   double *dst,
///   const double *src
/// ) const
/// This applies the transformation to a point.
/// @param[out] dst
///   The result is stored in this array of size 2.
/// @param[in] src
///   The original point is contained in this array of size 2.

/// @fn double EW_Transform2::get_denormalization() const
/// This calculates how far #orthog is from being orthogonal.
/// @return
///   The return value is a non-negative number representing this distance.

/// @fn void EW_Transform2::set_to_normalization(
///   const EW_Transform2 *a
/// )
/// This sets the transformation to another transformation and then coerces
/// #orthog to be orthogonal.
/// @param a
///   This points to the transformation to normalize.
///   @a a can equal @c this.

/// @fn bool EW_Transform2::get_valid() const
/// This checks that all components of #orthog, #scale and #translate are
/// finite floating point numbers and that #scale is positive and not
/// sub-normal.
/// @return
///   @c true if all these conditions are satisfied.

/// @fn bool EW_Transform2::operator==(const EW_Transform2 &a) const
/// This compares member by member, except for member #comps_cnt, which is
/// ignored.

/// @fn inline bool EW_Transform2::operator!=(const EW_Transform2 &a) const
/// This compares member by member, except for member #comps_cnt, which is
/// ignored.

/// @var const EW_Transform2 EW_Transform2::identity_transform
/// This is the identity transform.

// The rest of this file is private.

inline void
EW_Transform2::apply(float *dst, const float *src) const
{
  double f0 = translate[0] +
   scale * (orthog[0][0] * src[0] + orthog[0][1] * src[1]);
  double f1 = translate[1] +
   scale * (orthog[1][0] * src[0] + orthog[1][1] * src[1]);
  dst[0] = f0;
  dst[1] = f1;
}

inline void
EW_Transform2::apply(double *dst, const double *src) const
{
  double f0 = translate[0] +
   scale * (orthog[0][0] * src[0] + orthog[0][1] * src[1]);
  double f1 = translate[1] +
   scale * (orthog[1][0] * src[0] + orthog[1][1] * src[1]);
  dst[0] = f0;
  dst[1] = f1;
}

inline bool
EW_Transform2::operator!=(const EW_Transform2 &a) const
{
  return !operator==(a);
}

#endif
