// wdkg 2008-2010

#ifndef EW_TRANSFORM3_H_INCLUDED
#define EW_TRANSFORM3_H_INCLUDED 1

/// 3D Similarity Transformation

/// EW_Transform3 represents similarity transformations of @f$\mathbb{R}^3@f$,
/// each as the composition of an orthogonal linear map, a scaling and a
/// translation.
/// The translation is applied after the orthognal linear map and the scaling.
///
/// EW_Transform3 is a POD type class.
///
/// EW_Transform3 is declared in @c ew/transform3.h.

class EW_Transform3 {
public:
  double orthog[3][3];
  double scale;
  double translate[3];
  unsigned int comps_cnt;
  void set_to_identity();
  void set_to_inverse(
    const EW_Transform3 *a
  );
  void set_to_composition(
    const EW_Transform3 *a,
    const EW_Transform3 *b
  );
  bool set_to_interpolation(
    const EW_Transform3 *a,
    const EW_Transform3 *b,
    double e
  );
  void get_matrix_gl(
    double *buffer
  ) const;
  inline void apply(
    float *dst,
    const float *src
  ) const;
  inline void apply(
    double *dst,
    const double *src
  ) const;
  double get_denormalization() const;
  void set_to_normalization(
    const EW_Transform3 *a
  );
  bool get_valid() const;
  bool scan(const char **se, const char *s);
  bool format(char *buf, int bufl, int *outl) const;
  bool operator==(const EW_Transform3 &a) const;
  inline bool operator!=(const EW_Transform3 &a) const;
  static const int FORMAT_LEN = 480;
  static const EW_Transform3 identity_transform;
};

/// @var double EW_Transform3::orthog[3][3]
/// This is the matrix representing the orthogonal part of the transformation.
/// The matrix elements are ordered by row, and the matrix is applied on the
/// left.
/// The coordinates should be finite.
/// Deviations from orthogonaliity are tolerated.

/// @var double EW_Transform3::scale
/// This is the scale part of the transformation.
/// This should be positive.

/// @var double EW_Transform3::translate[3]
/// This is the translation part of the transformation.
/// The coordinates should be finite.

/// @var unsigned int EW_Transform3::comps_cnt
/// This is a count of the accumulated number of compositions performed in the
/// calculation of the current transformation.
/// It is reset by #set_to_identity and #set_to_normalization and set by
/// #set_to_inverse, #set_to_composition, #set_to_interpolation and #scan.
/// If the transformation is explicitly set, #comps_cnt should be too.
///
/// #set_to_composition will automatically normalize its result when #comps_cnt
/// exceeds an internal threshold.
///
/// This is to prevent exponential deviation of the orthogonal part from
/// orthogonality under iterated compositions.

/// @fn void EW_Transform3::set_to_identity()
/// This sets the transformation to the identity transformation.

/// @fn void EW_Transform3::set_to_inverse(
///   const EW_Transform3 *a
/// )
/// This sets the transformation to the inverse of another transformation.
/// @param a
///   This points to the transformation to invert.
///   @a a can equal @c this.

/// @fn void EW_Transform3::set_to_composition(
///   const EW_Transform3 *a,
///   const EW_Transform3 *b
/// )
/// This sets the transformation to the composition of two transformations.
/// @param a
///   This points to the transformation that is applied last in the
///   composition.
///   @a a can equal @c this.
/// @param b
///   This points to the transformation that is applied first in the
///   composition.
///   @a b can equal @c this.

/// @fn bool EW_Transform3::set_to_interpolation(
///   const EW_Transform3 *a,
///   const EW_Transform3 *b,
///   double e
/// )
/// This sets the transformation to an interpolation between two
/// transformations.
/// @param a
///   This points to the first transformation.
///   @a a can equal @c this.
/// @param b
///   This points to the second transformation.
///   @a b can equal @c this.
/// @param e
///   This is the interpolation paramater.
///   When 0.0, the transformation is set to @c *a, when 1.0 to @c *b.
///   @a e is not restricted to the range [0.0, 1.0].
/// @return
///   @c true if a smooth canonical interpolation was possible.
///   If @c false, the transformation is set to either the first or second
///   transformation, depending on whether @a e is closer to 0.0 or 1.0.
///   This can occur if one transformation is a reflection and the other
///   transformation is a rotation, or if the orthogonal parts differ by a half
///   turn.

/// @fn void EW_Transform3::get_matrix_gl(
///   double *buffer
/// ) const
/// This outputs the transformation as a 4 x 4 matrix in the format used by
/// @b OpenGL.
/// @param[out] buffer
///   The result is stored in this array of size 16.

/// @fn void EW_Transform3::apply(
///   float *dst,
///   const float *src
/// ) const
/// This applies the transformation to a point.
/// @param[out] dst
///   The result is stored in this array of size 3.
/// @param[in] src
///   The original point is contained in this array of size 3.

/// @fn void EW_Transform3::apply(
///   double *dst,
///   const double *src
/// ) const
/// This applies the transformation to a point.
/// @param[out] dst
///   The result is stored in this array of size 3.
/// @param[in] src
///   The original point is contained in this array of size 3.

/// @fn double EW_Transform3::get_denormalization() const
/// This calculates how far #orthog is from being orthogonal.
/// @return
///   The return value is a non-negative number representing this distance.

/// @fn void EW_Transform3::set_to_normalization(
///   const EW_Transform3 *a
/// )
/// This sets the transformation to another transformation and then coerces
/// #orthog to be orthogonal.
/// @param a
///   This points to the transformation to normalize.
///   @a a can equal @c this.

/// @fn bool EW_Transform3::get_valid() const
/// This checks that all components of #orthog, #scale and #translate are
/// finite floating point numbers and that #scale is positive and not
/// sub-normal.
/// @return
///   @c true if all these conditions are satisfied.

/// @fn bool EW_Transform3::scan(const char **se, const char *s)
/// This scans a text string for a representation of a transform.
/// Initial whitespace is not skipped, so would result in failure.
/// The string must be null terminated.
/// If successfully scanned, #comps_cnt is set to zero.
/// The transform is not automatically normalized, but can be manually.
/// @param[out] se
///   If non-zero, this is set to the address of the first unprocessed
///   character.
/// @param[in] s
///   This is a pointer to the text to scan.
/// @return
///   @c true if the initial segment of the text is a correctly formatted
///   transform.

/// @fn bool EW_Transform3::format(char *buf, int bufl, int *outl) const
/// This formats the transform as a text string.
///
/// The format used to represent transforms is as follows:
/// <CENTER>
/// [tr1 tr2 tr3] [or11 or12 or13 or21 or22 or23 or31 or32 or33] sc
/// </CENTER>
///
/// For example, this represents a transformation consisting of an enlargement
/// by a factor 1.5, followed by a rotation about the z axis, and then a
/// translation along the y axis:
/// <CENTER>
/// [0 4.5 0] [0.6 0.8 0 -0.8 0.6 0 0 0 1] 1.5
/// </CENTER>
///
/// The floating point numbers are represented exactly, but efficiently.
/// @param[out] buf
///   The output is written here.
/// @param bufl
///   The length of @a buf.
/// @param[out] outl
///   If non-null, the length of the output (even if it was not written),
///   not including the terminating null.
/// @return
///   @c true if the buffer is long enough for a successful format.

/// @fn bool EW_Transform3::operator==(const EW_Transform3 &a) const
/// This compares member by member, except for member #comps_cnt, which is
/// ignored.

/// @fn inline bool EW_Transform3::operator!=(const EW_Transform3 &a) const
/// This compares member by member, except for member #comps_cnt, which is
/// ignored.

/// @var const int EW_Transform3::FORMAT_LEN
/// This is the length of buffer that is guaranteed to be long enough for the
/// output of #format.

// This is 13 * EWText::FORMAT_DOUBLE_LEN + 64.
// Don't use EWText::FORMAT_DOUBLE_LEN though to keep Text.h internal.

/// @var const EW_Transform2 EW_Transform2::identity_transform
/// This is the identity transform.

// The rest of this file is private.

inline void
EW_Transform3::apply(float *dst, const float *src) const
{
  double f0 = translate[0] + scale *
   (orthog[0][0] * src[0] + orthog[0][1] * src[1] + orthog[0][2] * src[2]);
  double f1 = translate[1] + scale *
   (orthog[1][0] * src[0] + orthog[1][1] * src[1] + orthog[1][2] * src[2]);
  double f2 = translate[2] + scale *
   (orthog[2][0] * src[0] + orthog[2][1] * src[1] + orthog[2][2] * src[2]);
  dst[0] = f0;
  dst[1] = f1;
  dst[2] = f2;
}

inline void
EW_Transform3::apply(double *dst, const double *src) const
{
  double f0 = translate[0] + scale *
   (orthog[0][0] * src[0] + orthog[0][1] * src[1] + orthog[0][2] * src[2]);
  double f1 = translate[1] + scale *
   (orthog[1][0] * src[0] + orthog[1][1] * src[1] + orthog[1][2] * src[2]);
  double f2 = translate[2] + scale *
   (orthog[2][0] * src[0] + orthog[2][1] * src[1] + orthog[2][2] * src[2]);
  dst[0] = f0;
  dst[1] = f1;
  dst[2] = f2;
}

inline bool
EW_Transform3::operator!=(const EW_Transform3 &a) const
{
  return !operator==(a);
}

#endif
