// wdkg 2008-2010

#ifndef EW_VIEW_WIDGET_H_INCLUDED
#define EW_VIEW_WIDGET_H_INCLUDED 1

#include <stdexcept>
#include <string>
#include <vector>
#include "ew/dataflow_network.h"
#include "ew/dataflow_notifier.h"
#include "ew/transform3.h"

class EW_Bbox3;
class EW_View3_Item;
class EWDebug;

// XXX callback to activate context (not in destructors?)
// XXX no glErrors on calls
// XXXschedule_idle_handler exceptions passed thru to operations that attempt
// to change network
// XXX facility to facilitate automatically handling stacking order
// XXX ? use proj_to_sur or variant in pick

/// 3D viewing widget

/// This class implements a widget that manages a 3D scene consisting of a
/// list of items.
///
/// EW_View3_Widget is a class without assignment or comparison.
/// There are private member variables.
///
/// EW_View3_Widget is declared in @c ew/view3_widget.h.
///
/// Items, derived from EW_View3_Item, may be created in the widget.
///
/// This class is independent of any particular toolkit and it is intended that
/// the class will be inherited by a wrapper that implements the necessary
/// interface with the toolkit.
///
/// All view widgets associated with a particular EW_Dataflow_Network instance
/// must share OpenGL contexts.
/// The context must have a depth buffer.
/// Many class methods require this context to be active when they are called.
///
/// Items in a widget are kept in a list in rendering order.
/// If the depth buffer is not being used, later items will be rendered on top
/// of earlier items.
/// It the depth buffer is being used, the rendering order only effects items
/// with exactly the same depth.
/// Items are put at the end of the order initially on creation.
//XXX ?polygon offset

class EW_View3_Widget : private EW_Dataflow_Notifier {
public:
  explicit EW_View3_Widget(EW_Dataflow_Network *nw);
  virtual ~EW_View3_Widget();
  void reset();
  void set_clip_ratio(double cr);
  inline double get_clip_ratio() const;
  void set_view_mapping(const EW_Transform3 *tr);
  void view_post_compose(const EW_Transform3 *tr);
  inline const EW_Transform3 *get_view_mapping() const;
  void get_bbox(EW_Bbox3 *b) const;
  void get_pointer_location(double *loc, double x, double y, double z = 0.0)
   const;
  void get_center_location(double *loc) const;
  void get_pointer_translate_tr(EW_Transform3 *tr,
   double x, double y, double z) const;
  void get_pointer_roll_tr(EW_Transform3 *tr, double x, double y) const;
  void get_pointer_twist_tr(EW_Transform3 *tr, double x0, double y0,
   double x1, double y1, double acc) const;
  void get_pointer_scale_tr(EW_Transform3 *tr, double x0, double y0,
   double x1, double y1, double acc) const;
  void get_jump_tr(EW_Transform3 *tr, const double *loc, bool do_z) const;
  bool get_bbox_tr(EW_Transform3 *tr, const EW_Bbox3 *bbox, bool do_z,
   bool do_scale) const;
  void set_background_color(const unsigned char *rgb);
  inline const unsigned char *get_background_color() const;
  void set_highlight_color(const unsigned char *rgb);
  inline const unsigned char *get_highlight_color() const;
  inline EW_View3_Item *get_highlight_item() const;
  void clear_highlight();
  void set_use_depth(bool d);
  inline bool get_use_depth() const;
  void move_item(EW_View3_Item *it, int i);
  inline EW_View3_Item *const *get_items() const;
  inline int get_n_items() const;
  bool pick(double x, double y, double sz, double burrow,
   EW_View3_Item *constrain_it, int constrain_cmpt, int constrain_dim,
   EW_View3_Item **pick_it, int *pick_cmpt, int *pick_dim,
   double *pick_z);
  EW_Dataflow_Network *const network;
// XXX private to library ?eliminate
  inline bool get_has_double_buffer() const;
  inline bool get_has_stereo() const;
protected:
  void init_gl_context();
  void set_window_size(int w, int h);
  void render();
  void idle_handler();
  void set_currently_unrendered();
  void destruction_in_progress();
private:
  friend class EW_View3_Item;
// EW_View3_Item
  inline void redraw_later();
  void set_highlight_item(EW_View3_Item *it);
  inline void redraw_view_later();
  void add_item(EW_View3_Item *it);
  void del_item(EW_View3_Item *it);
  int dbg_new_item_num();
// private
  virtual void schedule_idle_handler_cb() = 0;
  virtual void redraw_cb() = 0;
  virtual void notifier_handle_cb();
  void dataflow_check();
  bool use_depth;
  bool has_double_buffer;
  bool has_stereo;
  bool currently_rendered;
  bool idle_pending;
  bool redraw_required;
  bool destruction_pending;
  EW_Dataflow_Network::cycle_t dataflow_check_cycle;
  double clip_ratio;
  unsigned char background_color[3];
  unsigned char highlight_color[3];
  EW_View3_Item *highlight_item;
  int winw;
  int winh;
  double projmatrix[16];
  EW_Transform3 view_mapping;
  EW_Transform3 window_mapping;
  std::vector<EW_View3_Item *> items;
  unsigned int DbgLastItemN;
public:
  const EWDebug &Dbg;
private:
  const bool DbgOn;
public:
// private to library
  const int DbgN;
private:
  const std::string DbgStr;
public:
// private to library
  const char *const DbgId;
// undefined
  EW_View3_Widget(const EW_View3_Widget &rhs);
  EW_View3_Widget &operator=(const EW_View3_Widget &rhs);
};

/// @fn EW_View3_Widget::EW_View3_Widget(EW_Dataflow_Network *nw)
/// This creates an empty view widget.
/// @param nw
///   A pointer to a network instance.
///   The view widget will be permanently associated with this network.

/// @fn EW_View3_Widget::~EW_View3_Widget()
/// This destroys the widget and all its items.
/// The OpenGL context should be active when this is called.

/// @fn void EW_View3_Widget::reset()
/// This deletes all items in the widget and resets it to its initial state.

/// @fn void EW_View3_Widget::set_clip_ratio(double cr)
/// The viewing volume width and height are determined by the window dimensions
/// as specified in #set_window_size.
/// The clipping ratio is the ratio of the viewing volume depth to the smaller
/// of its width and height.
/// The viewing volume coordinate system is centered in the viewing volume,
/// with axes, in order, pointing right, up and out of the screen, and with
/// homogeneous scale such that the range of coordinates in the direction of
/// the smaller of the width and height is [-1.0, 1.0].
/// The clip_ratio is considered a configuration setting.
/// It is initially 1.0.
/// @param cr
///   The new clipping ratio.
///   This must be positive, but also with unspecified limits on unreasonably
///   small or large values.

/// @fn double EW_View3_Widget::get_clip_ratio() const
/// @return
///   The current clipping ratio.

/// @fn void EW_View3_Widget::set_view_mapping(const EW_Transform3 *tr)
/// The view mapping is the transformation from model coordinates to viewing
/// volume coordinates.
/// @param[in] tr
///   The new view mapping.

/// @fn void EW_View3_Widget::view_post_compose(const EW_Transform3 *tr)
/// This composes the view mapping with a transform on the left.
/// This is for incremental changes to the view mapping relative to the viewing
/// volume coordinate system.
/// @param[in] tr
///   The view mapping increment.

/// @fn const EW_Transform3 *EW_View3_Widget::get_view_mapping() const
/// @return
///   A pointer to the current view mapping.

/// @fn void EW_View3_Widget::get_bbox(EW_Bbox3 *b) const
/// @param[out] b
///   Where to store the bounding box.

/// @fn void EW_View3_Widget::get_pointer_location(double *loc, double x,
///  double y, double z = 0.0) const
/// This returns the location of the point with given pixel window coordinates.
/// @param[out] loc
///   This points to an array of size 3 to store the location.
/// @param x, y, z
///   The pixel window coordinates (right from left side, down from top and
///   into window from the midplane of the viewing volume).

/// @fn void EW_View3_Widget::get_center_location(double *loc) const
/// This returns the location of center of the space currently in view.
/// @param[out] loc
///   This points to an array of size 3 to store the location.

/// @fn void EW_View3_Widget::get_pointer_translate_tr(EW_Transform3 *tr,
///  double x, double y, double z) const
/// This constructs a translation suitable for #view_post_compose from
/// window pixel coordinate changes.
/// @param[out] tr
///   The view mapping increment.
/// @param x, y, z
///   The coordinate changes.

/// @fn void EW_View3_Widget::get_pointer_roll_tr(EW_Transform3 *tr, double x,
///  double y) const
/// This constructs a roll-ball type rotation suitable for #view_post_compose
/// from window pixel coordinate changes.
/// @param[out] tr
///   The view mapping increment.
/// @param x, y
///   The coordinate changes.

/// @fn void EW_View3_Widget::get_pointer_twist_tr(EW_Transform3 *tr,
///  double x0, double y0, double x1, double y1, double acc) const
/// This constructs a twist rotation about the center of the window suitable
/// for #view_post_compose from window pixel coordinate changes.
/// @param[out] tr
///   The view mapping increment.
/// @param x0, y0
///   The coordinate initial values.
/// @param x1, y1
///   The coordinate final values.
/// @param acc
///   A acceleration factor that scales the rotation relative to the rotation
///   indicated by the coordinate changes.

/// @fn void EW_View3_Widget::get_pointer_scale_tr(EW_Transform3 *tr,
///  double x0, double y0, double x1, double y1, double acc) const
/// This constructs a scaling about the center of the window suitable
/// for #view_post_compose from window pixel coordinate changes.
/// @param[out] tr
///   The view mapping increment.
/// @param x0, y0
///   The coordinate initial values.
/// @param x1, y1
///   The coordinate final values.
/// @param acc
///   A acceleration factor that scales the scale relative to the scale
///   indicated by the coordinate changes.

/// @fn void EW_View3_Widget::get_jump_tr(EW_Transform3 *tr,
///  const double *loc, bool do_z) const
/// This constructs a translation suitable for #view_post_compose that
/// would put the specified point (in model space, not viewing space) at the
/// center of the view.
/// @param[out] tr
///   The view mapping increment.
/// @param[in] loc
///   This points to an array of size 3 containing the point coordinates
/// @param do_z
///   If @c true, translate so that the point is at the center of the view in
///   view x, y and z directions.
///   If @c false, don't translate in the view z direction.

/// @fn bool EW_View3_Widget::get_bbox_tr(EW_Transform3 *tr,
///  const EW_Bbox3 *bbox, bool do_z, bool do_scale) const
/// This constructs a translation suitable for #view_post_compose that
/// would put the specified bounding box (in model space, not viewing space) at
/// the center of the view.
/// The orientation of the view is not effected.
/// This will return @c false and not otherwise do anything if the bounding box
/// is empty, or if is a single point and @a do_scale is @c true.
/// @param[out] tr
///   The view mapping increment, if successful.
/// @param[in] bbox
///   The bounding box.
/// @param do_z
///   If @c true, translate so that the box is at the center of the view in
///   view x, y and z directions.
///   If @c false, don't translate in the view z direction.
/// @param do_scale
///   If @c true, scale so that the box fits comfortably in the view.
///   If @c false, don't scale.
/// @return
///   @c true if successful.

/// @fn void EW_View3_Widget::set_background_color(const unsigned char *rgb)
/// The background is by default [0, 0, 0].
/// @param[in] rgb
///   This points to an array of size 3 containing the new background colour.

/// @fn const unsigned char *EW_View3_Widget::get_background_color() const
/// @return
///   A pointer to the current background colour components.

/// @fn void EW_View3_Widget::set_highlight_color(const unsigned char *rgb)
/// The highlight colour is by default [255, 0, 0].
/// @param[in] rgb
///   This points to an array of size 3 containing the new highlight colour.

/// @fn const unsigned char *EW_View3_Widget::get_highlight_color() const
/// @return
///   A pointer to the current highlight colour components.

/// @fn EW_View3_Item *EW_View3_Widget::get_highlight_item() const
/// @return
///   A pointer to the current highlighted item, or 0.

/// @fn void EW_View3_Widget::clear_highlight()
/// This clears the current highlighted item, if any.

/// @fn void EW_View3_Widget::set_use_depth(bool d)
/// Turn on or off the use of the depth buffer for depth clipping.
/// It is initially @c true.
/// @param d
///   If @c true, use the depth buffer.

/// @fn bool EW_View3_Widget::get_use_depth() const
/// @return
///   The current setting.

/// @fn void EW_View3_Widget::move_item(EW_View3_Item *it, int n)
/// Move item in the rendering order.
/// @param it
///   A pointer to the item to move.
/// @param n
///   The position in the order to move the item to.

/// @fn inline EW_View3_Item *const *EW_View3_Widget::get_items() const
/// @return
///   A pointer to the list of items.

/// @fn inline int EW_View3_Widget::get_n_items() const
/// @return
///   The size of the list of items.

/// @fn bool EW_View3_Widget::pick(double x, double y, double sz,
///  double burrow, EW_View3_Item *constrain_it, int constrain_cmpt,
///  int constrain_dim, EW_View3_Item **pick_it, int *pick_cmpt, int *pick_dim,
///  double *pick_z)
/// This picks an item fragment rendered in a sub-window of the current view.
/// Fragments of the lowest dimension visible are picked.
/// With a depth buffer, the nearest eligible item fragment of this dimensions
/// is picked, provided it is not behind, with a tolerance, any surface
/// fragment.
/// This is to allow picking of, for example, points in an inclined surface
/// where otherwise, some part of the surface would always be closer than the
/// point.
/// Without a depth buffer, the latest eligible item fragment of this dimension
/// in the rendering order is picked, provided it is not before any surface
/// fragment.
/// @param[in] x, y
///   The pixel window coordinates.
/// @param[in] sz
///   The side of the square, in pixels, centered on @a x and @a y within which
///   rendered items are eligible for picking.
/// @param[in] burrow
///   The tolerance in the depth buffer, for an eligible item to be picked
///   behind a surface.
/// @param[in] constrain_it
///   If this is non-zero, only fragments rendered by this item will be
///   eligible.
/// @param[in] constrain_cmpt
///   If this is non-negative, only fragments rendered as part of this
///   component in the item will be eligible.
///   The meaning of the component depends on the item.
/// @param[in] constrain_dim
///   If this is non-negative, only fragments of this dimension (0, 1 or 2)
///   will be eligible.
/// @param[out] pick_it
///   If non-zero and the pick was successful, a pointer to the picked item is
///   stored here.
/// @param[out] pick_cmpt
///   If non-zero and the pick was successful, the component of the picked item
///   is stored here.
///   The meaning of the component depends on the item.
/// @param[out] pick_dim
///   If non-zero and the pick was successful, the dimension of the picked
///   fragment is stored here.
/// @param[out] pick_z
///   If non-zero and the pick was successful, the z coordinate of the nearest
///   part of the picked item part is stored here.
///   This is in pixels, with 0 at the center of the viewing volume, and
///   positive further away.
///   It is suitable for use in #get_pointer_location.

/// @var EW_Dataflow_Network *const EW_View3_Widget::network
/// This points to the EW_Dataflow_Network the view widget is associated with.

/// @fn void EW_View3_Widget::init_gl_context()
/// This performs initialisation related to the OpenGL context.
/// It should be called after construction, once the context is active.

/// @fn void EW_View3_Widget::set_window_size(int w, int h)
/// This performs initialisation related to the window size.
/// It should be called after construction and whenever the window is resized.
/// @param w, h
///   The new window size.

/// @fn void EW_View3_Widget::render()
/// This renders the scene in the current OpenGL context and associated window.
/// If the context has a double buffer, this will render to the back buffer.

/// @fn void EW_View3_Widget::idle_handler()
/// This should be called at an appropriate time when requested by
/// #schedule_idle_handler_cb.

/// @fn void EW_View3_Widget::set_currently_unrendered()
/// This should be called when the window associated with this widget is
/// unmapped, to avoid unnecessary subsequent activity.
/// A call to #render undoes the effect of this.

/// @fn void EW_View3_Widget::destruction_in_progress()
/// This should be called when a derived class instance is being destroyed, to
/// prevent any callbacks being invoked.

/// @fn void EW_View3_Widget::schedule_idle_handler_cb()
/// This is a callback that must be implemented by the inheriting class.
/// It must arrange for #idle_handler to be called later when the GUI is idle.
/// Outstanding handlers must be cancelled by the inheriting class on
/// destruction.

/// @fn void EW_View3_Widget::redraw_cb()
/// This is a callback than must be implemented in the inheriting class.
/// It must arrange for the widget to be rendered immediately.

// The rest of this file is private.

inline int
EW_View3_Widget::get_n_items() const
{
  return items.size();
}

inline EW_View3_Item *const *
EW_View3_Widget::get_items() const
{
  return items.size() == 0 ? 0 : &items[0];
}

inline bool
EW_View3_Widget::get_has_double_buffer() const
{
  return has_double_buffer;
}

inline bool
EW_View3_Widget::get_has_stereo() const
{
  return has_stereo;
}

inline double
EW_View3_Widget::get_clip_ratio() const
{
  return clip_ratio;
}

inline const unsigned char *
EW_View3_Widget::get_background_color() const
{
  return background_color;
}

inline const unsigned char *
EW_View3_Widget::get_highlight_color() const
{
  return highlight_color;
}

inline EW_View3_Item *
EW_View3_Widget::get_highlight_item() const
{
  return highlight_item;
}

inline const EW_Transform3 *
EW_View3_Widget::get_view_mapping() const
{
  return &view_mapping;
}

inline void
EW_View3_Widget::redraw_later()
{
  if (currently_rendered && !idle_pending && !destruction_pending) {
    schedule_idle_handler_cb();
    idle_pending = true;
  }
  redraw_required = true;
}

inline bool
EW_View3_Widget::get_use_depth() const
{
  return use_depth;
}

#endif
